// *****************************************************************************
// NOTES:  This code is for editing existing geodatabases, tables, etc.,
//         not creating new ones if they do not already exist.
// *****************************************************************************

#include <CEditSQLiteDb.h>

#include <iostream>

#include <CppSQLite3.h>
#include <TableDef.h>

class CEditSQLiteDb::impl
{
public:
  impl(CppSQLite3DB* a_) : m_db(a_) {}


  int  SetupNewGDB();
  int  CreateTable(CComBSTR a_TableName);
  bool TableExists(CComBSTR a_TableName);
  int  EditTable(CComBSTR a_TableName,
                 CComSafeArray<BSTR> &a_FieldNames,
                 CComSafeArray<VARIANT> &a_2DValues);
  int  EditExistingTableOneRow(CComBSTR a_TableName,
                               CComSafeArray<BSTR> &a_FieldNames,
                               CComSafeArray<VARIANT> &a_2DValues,
                               std::string a_query);
  int  EditTableArrayDat(CComBSTR a_TableName,
                         CComSafeArray<BSTR> &a_FieldNames,
                         CComSafeArray<VARIANT> &a_2DValues);
  int  EditExistingTableMultipleRows(CComBSTR a_TableName,
                                     CComSafeArray<BSTR> &a_FieldNames,
                                     CComSafeArray<VARIANT> &a_2DValues,
                                     std::string a_query,
                                     int &a_nNumRowsEdited);
  int DeleteTable(CComBSTR a_TableName);

private:
  impl(const impl&);
  const impl& operator=(const impl&);

  CppSQLite3DB* m_db;
};

std::string iGetTypeString (esriFieldType a_type)
{
  if (a_type == esriFieldTypeInteger
      || a_type == esriFieldTypeSmallInteger
      )
  {
    return "INTEGER";
  }
  else if (a_type == esriFieldTypeDouble)
  {
    return "DOUBLE";
  }
  else
  {
    return "TEXT";
  }
} // iGetTypeString
//-----------------------------------------------------------------------------
/// \brief Constructor
//-----------------------------------------------------------------------------
CEditSQLiteDb::CEditSQLiteDb (CppSQLite3DB* a_) :
m_p(new CEditSQLiteDb::impl(a_))
{
} // CEditSQLiteDb::CEditSQLiteDb
//-----------------------------------------------------------------------------
/// \brief Destructor
//-----------------------------------------------------------------------------
CEditSQLiteDb::~CEditSQLiteDb ()
{
  try
  {
    if (m_p)
      delete(m_p);
  }
  catch (...) {}
} // CEditSQLiteDb:~CEditSQLiteDb
//-----------------------------------------------------------------------------
/// \brief Sets up a new GeoDb and creates certain tables by default
//-----------------------------------------------------------------------------
int CEditSQLiteDb::SetupNewGDB ()
{
  return m_p->SetupNewGDB();
} // CEditSQLiteDb::SetupNewGDB
//-----------------------------------------------------------------------------
/// \brief Creates and edits a table
//-----------------------------------------------------------------------------
int CEditSQLiteDb::EditTable (CComBSTR a_TableName,
                              CComSafeArray<BSTR> &a_FieldNames,
                              CComSafeArray<VARIANT> &a_2DValues)
{
  if (!m_p->TableExists(a_TableName))
    m_p->CreateTable(a_TableName);

  m_p->EditTable(a_TableName,
                 a_FieldNames,
                 a_2DValues);

  return 1;
} // CEditSQLiteDb::EditTable
//-----------------------------------------------------------------------------
/// \brief Edits a table
//-----------------------------------------------------------------------------
int CEditSQLiteDb::EditTableExistingRows (CComBSTR a_TableName,
                                          CComSafeArray<BSTR>& a_FieldNames,
                                          CComSafeArray<VARIANT>& a_2DValues,
                                          int &a_nNumRowsEdited,
                                          CComBSTR a_WhereClause /*= ""*/)
{
  if (!m_p->TableExists(a_TableName))
    return 1;

  _bstr_t strIntermediate(a_WhereClause);
  CStr query(strIntermediate);
  std::string sStr = (LPCTSTR)strIntermediate;
  if (query.find(" = ") != -1)
  {
    if (m_p->EditExistingTableOneRow(a_TableName, a_FieldNames, a_2DValues, sStr) > 0)
      a_nNumRowsEdited = 1;
  }
  else if (query.find("<") != -1 &&
           query.find(">") != -1)
  {
    m_p->EditExistingTableMultipleRows(a_TableName,
                                       a_FieldNames,
                                       a_2DValues,
                                       sStr,
                                       a_nNumRowsEdited);
  }
  else
  {
    ASSERT(0);
  }
  return 1;
} // CEditSQLiteDb::EditTableExistingRows
//-----------------------------------------------------------------------------
/// \brief Edits a table
//-----------------------------------------------------------------------------
int CEditSQLiteDb::DeleteTable (CComBSTR a_TableName)
{
  if (m_p->TableExists(a_TableName))
    return m_p->DeleteTable(a_TableName);
  else
    return 1;
} // CEditSQLiteDb::DeleteTable
//-----------------------------------------------------------------------------
/// \brief Creates some empty tables in the sqlite db
//-----------------------------------------------------------------------------
int CEditSQLiteDb::impl::SetupNewGDB ()
{
  // output tables
  CreateTable("OutputFiles");
  CreateTable("OutputTime");
  CreateTable("OutputHead");
  CreateTable("OutputDrawdown");
  CreateTable("OutputFlow");
  // param tables
  CreateTable("Params");
  CreateTable("Clusters");
  CreateTable("IZ");
  CreateTable("PilotPts");
  CreateTable("Multipliers");
  CreateTable("MultNames");
  CreateTable("Zones");
  CreateTable("ZoneNames");
  CreateTable("ParInstances");
  CreateTable("ParInstSP");
  // Observation tables
  CreateTable("OBSVars");
  CreateTable("HOB");
  CreateTable("HOBLayers");
  CreateTable("HOBTimes");
  CreateTable("FLOB");
  CreateTable("FLOBFactors");
  return 1;
} // CEditSQLiteDb::impl::SetupNewGDB
//-----------------------------------------------------------------------------
/// \brief Creates a table
//-----------------------------------------------------------------------------
bool CEditSQLiteDb::impl::TableExists (CComBSTR a_TableName)
{
  bool rval(false);

  _bstr_t strIntermediate(a_TableName);
  std::string tab((LPCTSTR)strIntermediate);

  if (m_db->tableExists(tab.c_str()))
    rval = true;
  return rval;
} // CEditSQLiteDb::impl::TableExists
//-----------------------------------------------------------------------------
/// \brief Creates a table
//-----------------------------------------------------------------------------
int CEditSQLiteDb::impl::CreateTable (CComBSTR a_TableName)
{
  CTableDef tableDef;
  int tabId(tableDef.GetTableIdFromString(a_TableName.m_str));

  CComSafeArray<BSTR> colNames;
  VEC_FTYPE colTypes;
  VEC_DOMAIN colDomains;
  CComBSTR tableName;
  tableDef.GetData(tabId, colNames, colTypes, colDomains, tableName);

  _bstr_t strIntermediate;
  std::string dmlStr("create table ");
  strIntermediate = tableName;
  dmlStr += (LPCTSTR)strIntermediate;
  dmlStr += "(OID INTEGER PRIMARY KEY AUTOINCREMENT, ";
  for (ULONG i=0; i<colNames.GetCount(); i++)
  {
    if (i > 0)
      dmlStr += ", ";
    strIntermediate = colNames.GetAt(i);
    dmlStr += "'";
    dmlStr += (LPCTSTR)strIntermediate;
    dmlStr += "' ";
    dmlStr += iGetTypeString(colTypes.at(i));
  }
  dmlStr += ");";

  //m_db->execDML("begin transaction;");
  m_db->execDML(dmlStr.c_str());
  //m_db->execDML("commit transaction;");

  return 1;
} // CEditSQLiteDb::impl::CreateTable
//-----------------------------------------------------------------------------
/// \brief Creates a table
//-----------------------------------------------------------------------------
int CEditSQLiteDb::impl::EditTable (CComBSTR a_TableName,
                                    CComSafeArray<BSTR> &a_FieldNames,
                                    CComSafeArray<VARIANT> &a_2DValues)
{
  // see if it is only 1 column and array data
  if (EditTableArrayDat(a_TableName, a_FieldNames, a_2DValues))
    return 1;

  //m_db->execDML("begin transaction;");
  try
  {

    ULONG nRow, nCol;
    nCol = a_FieldNames.GetCount();
    nRow = a_2DValues.GetCount(1);

    VARIANT v;
    _variant_t varT;
    CComBSTR bStr;
    _bstr_t strInter;
    long idx[2];
    for (ULONG i=0; i<nRow; i++)
    {
      strInter = a_TableName;
      std::string dmlStr;
      dmlStr = "insert into ";
      dmlStr += (LPCTSTR)strInter;
      dmlStr += " (";
      for (ULONG j=0; j<nCol; j++)
      {
        if (j > 0)
          dmlStr += ", ";
        strInter = a_FieldNames.GetAt(j);
        dmlStr += (LPCTSTR)strInter;
      }
      dmlStr += ")";

      dmlStr += " values (";

      idx[1] = i;
      for (ULONG j=0; j<nCol; j++)
      {
        if (j > 0)
          dmlStr += ", ";
        idx[0] = j;
        a_2DValues.MultiDimGetAt(idx, v);
        varT = v;
        strInter = varT;
        if (v.vt == VT_BSTR)
        {
          dmlStr += "'";
        }
        dmlStr += (LPCTSTR)strInter;
        if (v.vt == VT_BSTR)
        {
          dmlStr += "'";
        }
      }
      dmlStr += ");";
      m_db->execDML(dmlStr.c_str());
    }

  }
  catch (CppSQLite3Exception& e)
  {
    std::cout << e.errorCode() << ":" << e.errorMessage() << std::endl;
  }
  //m_db->execDML("commit transaction;");
  return 1;
} // CEditSQLiteDb::impl::EditTable
//-----------------------------------------------------------------------------
/// \brief Creates a table
//-----------------------------------------------------------------------------
int CEditSQLiteDb::impl::EditTableArrayDat (CComBSTR a_TableName,
                                            CComSafeArray<BSTR> &a_FieldNames,
                                            CComSafeArray<VARIANT> &a_2DValues)
{
  if (a_FieldNames.GetCount() != 2 ||
      a_2DValues.GetCount(0) > 2)
  {
    return 0;
  }

  VARIANT v;
  _variant_t varT;
  long idx[2] = {0,0};
  a_2DValues.MultiDimGetAt(idx, v);
  varT = v;
  if (v.vt != VT_R8 &&
      v.vt != VT_R4)
    return 0;

  ULONG nRows(a_2DValues.GetCount(1));
  _bstr_t strInter(a_TableName);
  std::string dmlStr("insert into ");
  dmlStr += (LPCTSTR)strInter;
  dmlStr += " (";
  strInter = a_FieldNames.GetAt(0);
  dmlStr += (LPCSTR)strInter;
  dmlStr += ", ";
  strInter = a_FieldNames.GetAt(1);
  dmlStr += (LPCSTR)strInter;
  dmlStr += ") values(?, ?);";

  //m_db->execDML("begin transaction;");
  CppSQLite3Statement stmt = m_db->compileStatement(dmlStr.c_str());
  int ijk;
  double d;
  for (ULONG i=0; i<nRows; i++)
  {
    idx[0] = 1;
    idx[1] = i;
    a_2DValues.MultiDimGetAt(idx, v);
    varT = v;
    ijk = (int)varT;

    idx[0] = 0;
    idx[1] = i;
    a_2DValues.MultiDimGetAt(idx, v);
    varT = v;
    d = (double)varT;
    stmt.bind(1, d);
    stmt.bind(2, ijk);
    stmt.execDML();
    stmt.reset();
  }
  //m_db->execDML("commit transaction;");
  return 1;
} // CEditSQLiteDb::impl::EditTableArrayDat
//-----------------------------------------------------------------------------
/// \brief Creates a table
//-----------------------------------------------------------------------------
int CEditSQLiteDb::impl::EditExistingTableOneRow (CComBSTR a_TableName,
                                                  CComSafeArray<BSTR> &a_FieldNames,
                                                  CComSafeArray<VARIANT> &a_2DValues,
                                                  std::string a_query)
{
  int rval(1);
  //m_db->execDML("begin transaction;");
  try
  {

    ULONG nRow, nCol;
    nCol = a_FieldNames.GetCount();
    nRow = a_2DValues.GetCount(1);

    VARIANT v;
    CComBSTR bStr;
    long idx[2];
    _bstr_t strInter;
    _variant_t varT;
    for (ULONG i=0; i<nRow; i++)
    {
      strInter = a_TableName;
      std::string dmlStr;
      dmlStr = "UPDATE ";
      dmlStr += (LPCTSTR)strInter;
      dmlStr += " SET ";
      idx[1] = i;
      for (ULONG j=0; j<nCol; j++)
      {
        if (j > 0)
          dmlStr += ", ";
        strInter = a_FieldNames.GetAt(j);
        dmlStr += (LPCTSTR)strInter;
        dmlStr += " = ";

        idx[0] = j;
        a_2DValues.MultiDimGetAt(idx, v);
        varT = v;
        strInter = varT;
        if (v.vt == VT_BSTR)
        {
          dmlStr += "'";
        }
        dmlStr += (LPCTSTR)strInter;
        if (v.vt == VT_BSTR)
        {
          dmlStr += "'";
        }
      }

      dmlStr += " WHERE ";
      dmlStr += a_query;
      dmlStr += ";";
      int nRow = m_db->execDML(dmlStr.c_str());
      if (nRow < 1)
        rval = 0;
    }

  }
  catch (CppSQLite3Exception& e)
  {
    std::cout << e.errorCode() << ":" << e.errorMessage() << std::endl;
  }
  //m_db->execDML("commit transaction;");
  return rval;
} // CEditSQLiteDb::impl::EditExistingTableOneRow
//-----------------------------------------------------------------------------
/// \brief Creates a table
//-----------------------------------------------------------------------------
int CEditSQLiteDb::impl::EditExistingTableMultipleRows (CComBSTR a_TableName,
                                                        CComSafeArray<BSTR> &a_FieldNames,
                                                        CComSafeArray<VARIANT> &a_2DValues,
                                                        std::string a_query,
                                                        int &a_nNumRowsEdited)
{
  _bstr_t strInter;
  // get the starting index and the ending index from the query string
  int startIdx(0);
  CStr str(a_query.c_str()), strRight, strLeft;
  strRight = str.Right(str.GetLength()-str.Find(">")-2);
  strLeft = strRight.Left(strRight.Find(" "));
  std::string strInt((LPCTSTR)strLeft);
  startIdx = atoi(strInt.c_str());
  startIdx++;

  std::string queryField(str.Left(str.Find(" ")));
  if (a_FieldNames.GetCount() > 1)
  {
    strInter = a_FieldNames.GetAt(1);
    queryField = (LPCTSTR)strInter;
  }

  a_nNumRowsEdited = 0;
  // see if the items are in the table
  {
    std::string dmlStr("select count(*) from ");
    strInter = a_TableName;
    dmlStr += (LPCTSTR)strInter;
    dmlStr += " WHERE ";
    dmlStr += a_query;
    dmlStr += ";";
    if (m_db->execScalar(dmlStr.c_str()) < 1)
      return 1;
  }

  try
  {
    ULONG nRow, nCol;
    nCol = a_FieldNames.GetCount();
    nRow = a_2DValues.GetCount(1);


    strInter = a_TableName;
    std::string dmlStr;
    dmlStr = "UPDATE ";
    dmlStr += (LPCTSTR)strInter;
    dmlStr += " SET ";
    strInter = a_FieldNames.GetAt(0);
    dmlStr += (LPCTSTR)strInter;
    dmlStr += " = ? WHERE ";
    dmlStr += queryField;
    dmlStr += " = ?;";

    std::string idxStr("CREATE INDEX idx_tmp on ");
    strInter = a_TableName;
    idxStr += (LPCTSTR)strInter;
    idxStr += " (";
    idxStr += queryField;
    idxStr += ");";

    VARIANT v;
    CComBSTR bStr;
    long idx[2];
    _variant_t varT;
    double d;

    //m_db->execDML("begin transaction;");
    m_db->execDML(idxStr.c_str());
    CppSQLite3Statement stmt = m_db->compileStatement(dmlStr.c_str());
    for (ULONG i=0; i<nRow; i++, startIdx++, a_nNumRowsEdited++)
    {
      idx[0] = 0;
      idx[1] = i;
      a_2DValues.MultiDimGetAt(idx, v);
      varT = v;
      d = varT;
      stmt.bind(1, d);
      stmt.bind(2, startIdx);
      stmt.execDML();
      stmt.reset();
    }
    m_db->execDML("DROP INDEX idx_tmp;");
    //m_db->execDML("commit transaction;");
  }
  catch (CppSQLite3Exception& e)
  {
    std::cout << e.errorCode() << ":" << e.errorMessage() << std::endl;
  }
  return 1;
} // CEditSQLiteDb::impl::EditExistingTableMultipleRows
//-----------------------------------------------------------------------------
/// \brief Deletes all the rows in a table
//-----------------------------------------------------------------------------
int CEditSQLiteDb::impl::DeleteTable (CComBSTR a_TableName)
{
  int rval(1);
  _bstr_t strInter;
  try
  {
    std::string dmlStr("Drop Table ");
    strInter = a_TableName;
    dmlStr += (LPCTSTR)strInter;
    dmlStr += ";";

    //m_db->execDML("begin transaction;");
    m_db->execDML(dmlStr.c_str());
    //m_db->execDML("commit transaction;");
  }
  catch (CppSQLite3Exception& e)
  {
    std::cout << e.errorCode() << ":" << e.errorMessage() << std::endl;
  }
  return 1;
} // CEditSQLiteDb::impl::DeleteTable

///////////////////////////////////////////////////////////////////////////////
// TESTS
///////////////////////////////////////////////////////////////////////////////
#if CXX_TEST
#include <CEditSQLiteDb.t.h>
#include <CppSQLite3.h>


//-----------------------------------------------------------------------------
void CEditSQLiteDbTester::testCreateClass ()
{
  CEditSQLiteDb *p(0);
  p = new CEditSQLiteDb(0);
  if (!p)
    TS_FAIL("");
  else
    delete(p);
}
//-----------------------------------------------------------------------------
void CEditSQLiteDbTester::testTableExists ()
{
  CppSQLite3DB db;
  db.open(":memory:");
  {
    CEditSQLiteDb sq(&db);
    TS_ASSERT(!sq.m_p->TableExists("StressPeriods"));
    sq.m_p->CreateTable("StressPeriods");
    TS_ASSERT(sq.m_p->TableExists("StressPeriods"));
  }
  db.close();
}
//-----------------------------------------------------------------------------
void CEditSQLiteDbTester::testCreateTable ()
{
  //char c[5000];
  //GetModuleFileName(NULL, c, 5000);
  //CStr str(c);
  //int pos(str.ReverseFind("\\"));
  //CStr str1 = str.Left(str.GetLength() - (str.GetLength() - pos));
  //str1 += "\\test.db";
  //remove(str1.c_str());

  {
    CppSQLite3DB db;
    db.open(":memory:");
    {
      CEditSQLiteDb sq(&db);
      sq.m_p->CreateTable("StressPeriods");
      TS_ASSERT(db.tableExists("StressPeriods"));
    }
    db.close();
  }
}
//-----------------------------------------------------------------------------
void CEditSQLiteDbTester::testEditTable ()
{
  {
    CppSQLite3DB db;
    db.open(":memory:");
    {
      CEditSQLiteDb sq(&db);
      CComBSTR tab("StressPeriods");
      sq.m_p->CreateTable(tab);

      // insert into the table
//int CEditSQLiteDb::impl::EditTable (CComBSTR a_TableName,
//                                    CComSafeArray<BSTR> &a_FieldNames,
//                                    CComSafeArray<VARIANT> &a_2DValues)
      CComSafeArray<BSTR> colNames;
      CComSafeArray<VARIANT> vals2D;
      {
        CTableDef tableDef;
        int tabId(tableDef.GetTableIdFromString(tab.m_str));
        VEC_FTYPE colTypes;
        VEC_DOMAIN colDomains;
        CComBSTR tableName;
        tableDef.GetData(tabId, colNames, colTypes, colDomains, tableName);
      }

      {
        CComSafeArrayBound bound[2];
        bound[0].SetLowerBound(0); bound[0].SetCount(colNames.GetCount());
        bound[1].SetLowerBound(0); bound[1].SetCount(2);
        vals2D.Create(bound, 2);

        long idx[2]={0,0};
        for (int j = 0; j<2; j++)
        {
          idx[0] = 0;
          idx[1] = j;
          vals2D.MultiDimSetAt(idx, CComVariant(2)); // SPID
          idx[0] = 1;
          vals2D.MultiDimSetAt(idx, CComVariant(25.2)); // PERLEN
          idx[0] = 2;
          vals2D.MultiDimSetAt(idx, CComVariant(3)); // NSTP
          idx[0] = 3;
          vals2D.MultiDimSetAt(idx, CComVariant(1.5)); // TSMULT
          idx[0] = 4;
          vals2D.MultiDimSetAt(idx, CComVariant("SS")); // SsorTr

          for (int i = 0; i<11; i++)
          {
            idx[0] = 5+i;
            vals2D.MultiDimSetAt(idx, CComVariant(40+i)); // RIV_ITMP to EVT_INIEVT
          }
        }
      }
      sq.m_p->EditTable(tab, colNames, vals2D);
      {
        CComSafeArrayBound bound[2];
        bound[0].SetLowerBound(0); bound[0].SetCount(colNames.GetCount());
        bound[1].SetLowerBound(0); bound[1].SetCount(1);
        vals2D.Destroy();
        vals2D.Create(bound, 2);
        long idx[2]={0,0};
        for (int j = 0; j<1; j++)
        {
          idx[0] = 0;
          idx[1] = j;
          vals2D.MultiDimSetAt(idx, CComVariant(3)); // SPID
          idx[0] = 1;
          vals2D.MultiDimSetAt(idx, CComVariant(25.3)); // PERLEN
          idx[0] = 2;
          vals2D.MultiDimSetAt(idx, CComVariant(4)); // NSTP
          idx[0] = 3;
          vals2D.MultiDimSetAt(idx, CComVariant(1.6)); // TSMULT
          idx[0] = 4;
          vals2D.MultiDimSetAt(idx, CComVariant("TR")); // SsorTr

          for (int i = 0; i<11; i++)
          {
            idx[0] = 5+i;
            vals2D.MultiDimSetAt(idx, CComVariant(400+i)); // RIV_ITMP to EVT_INIEVT
          }
        }
      }
      sq.m_p->EditExistingTableOneRow(tab, colNames, vals2D, "OID = 2");
    }
    db.close();
  }
}
//-----------------------------------------------------------------------------
void CEditSQLiteDbTester::testEditTableArrayDat ()
{
  {
    CppSQLite3DB db;
    db.open(":memory:");
    {
      CEditSQLiteDb sq(&db);
      CComBSTR tab("LPFProperties");
      sq.m_p->CreateTable(tab);

      CComSafeArray<BSTR> colNames;
      CComSafeArray<VARIANT> vals2D;

      colNames.Add(CComBSTR("HK"));
      colNames.Add(CComBSTR("IJK"));

      {
        CComSafeArrayBound bound[2];
        bound[0].SetLowerBound(0); bound[0].SetCount(colNames.GetCount());
        bound[1].SetLowerBound(0); bound[1].SetCount(100);
        vals2D.Create(bound, 2);
        long idx[2]={0,0};
        for (int i=0; i<100; i++)
        {
          idx[0] = 0;
          idx[1] = i;
          double d((i+1)*1.5);
          vals2D.MultiDimSetAt(idx, CComVariant(d));
          idx[0] = 1;
          vals2D.MultiDimSetAt(idx, CComVariant(i+1));
        }
      }
      sq.m_p->EditTableArrayDat(tab, colNames, vals2D);
      colNames.SetAt(0, CComBSTR("HANI"));
      int nRow;
      sq.m_p->EditExistingTableMultipleRows(tab, colNames, vals2D,
                                            "ijk > 0 and ijk < 101",
                                            nRow);

    }
    db.close();
  }
}


#endif

