//------------------------------------------------------------------------------
// FILE      DataToSafeArray.cpp
// PURPOSE   
// COPYRIGHT Aquaveo 2006 All rights reserved.
//------------------------------------------------------------------------------

#include <DataToSafeArray.h>

#include <map>

#include <Grid.h>
#include <TableDef.h>


//------------------------------------------------------------------------------
/// \brief This returns the ID associated with a cell. For right now we just
/// return what was passed in, but this will make it easier to change to using
/// hydroIds.
//------------------------------------------------------------------------------
static int CellId (int a_modflowCellId)
{
  return a_modflowCellId;
} // CellId
//------------------------------------------------------------------------------
/// \brief This packages up array data coming from U2DREL
//------------------------------------------------------------------------------
template <class T>
int DataToSafeArray::OnlyArrayU2DTemplate (const char *a_table,
                                           const char *a_field,
                                           int a_nData,
                                           const T *a_data,
                                           CComSafeArray<VARIANT> &a_vals,
                                           CComSafeArray<BSTR> &a_fieldNames,
                                           CComBSTR &a_tableName)
{
  if (a_nData < 1 || !a_data)
    return 0;

  // set the table name and field name
  a_tableName = a_table;
  a_fieldNames.Add(CComBSTR(a_field));

  // size the VARIANT array
  CComSafeArrayBound bound[2];
  bound[0].SetLowerBound(0); bound[0].SetCount(a_fieldNames.GetCount());
  bound[1].SetLowerBound(0); bound[1].SetCount(a_nData);
  a_vals.Create(bound, 2);
  long idx[2] = {0,0};

  for (int i=0; i<a_nData; i++)
  {
    idx[0] = 0; idx[1] = i;
    a_vals.MultiDimSetAt(idx, CComVariant(a_data[i]));
  }

  return 1;
} // DataToSafeArray::OnlyArrayU2DTemplate
// explicit instantiation
int DataToSafeArray::OnlyArrayU2DINT (const char *a_table,
                                      const char *a_field,
                                      int a_nData,
                                      const int *a_data,
                                      CComSafeArray<VARIANT> &a_vals,
                                      CComSafeArray<BSTR> &a_fieldNames,
                                      CComBSTR &a_tableName)
{
  return (OnlyArrayU2DTemplate(a_table, a_field, a_nData, a_data, a_vals,
                               a_fieldNames, a_tableName));
} // DataToSafeArray::OnlyArrayU2DINT
int DataToSafeArray::OnlyArrayU2DREL (const char *a_table,
                                      const char *a_field,
                                      int a_nData,
                                      const float *a_data,
                                      CComSafeArray<VARIANT> &a_vals,
                                      CComSafeArray<BSTR> &a_fieldNames,
                                      CComBSTR &a_tableName)
{
  return (OnlyArrayU2DTemplate(a_table, a_field, a_nData, a_data, a_vals,
                               a_fieldNames, a_tableName));
} // DataToSafeArray::OnlyArrayU2DREL
int DataToSafeArray::OnlyArrayU2DREL (const char *a_table,
                                      const char *a_field,
                                      int a_nData,
                                      const double *a_data,
                                      CComSafeArray<VARIANT> &a_vals,
                                      CComSafeArray<BSTR> &a_fieldNames,
                                      CComBSTR &a_tableName)
{
  return (OnlyArrayU2DTemplate(a_table, a_field, a_nData, a_data, a_vals,
                               a_fieldNames, a_tableName));
} // DataToSafeArray::OnlyArrayU2DREL
//------------------------------------------------------------------------------
/// \brief This packages up array data coming from U2DREL
//------------------------------------------------------------------------------
template <class T>
int DataToSafeArray::U2DTemplate (const char *a_table,
                                  const char *a_cellField,
                                  const char *a_field,
                                  int a_nData,
                                  int a_startCellId,
                                  int a_stressPeriod,
                                  int a_etsSegId,
                                  const T *a_data,
                                  CComSafeArray<VARIANT> &a_vals,
                                  CComSafeArray<BSTR> &a_fieldNames,
                                  CComBSTR &a_tableName)
{
  if (!a_data || a_nData < 1 || a_startCellId < 1)
    return 0;

  a_tableName = a_table;
  // make the field names
  a_fieldNames.Add(CComBSTR(a_field));
  a_fieldNames.Add(CComBSTR(a_cellField));
  if (a_stressPeriod > 0)
  {
    if (a_tableName == "HUFUnitProps")
      a_fieldNames.Add(CComBSTR("HGUID"));
    else
      a_fieldNames.Add(CComBSTR("SPID"));
    if (a_etsSegId > 0)
      a_fieldNames.Add(CComBSTR("SEGID"));
  }

  // size the VARIANT array
  CComSafeArrayBound bound[2];
  bound[0].SetLowerBound(0); bound[0].SetCount(a_fieldNames.GetCount());
  bound[1].SetLowerBound(0); bound[1].SetCount(a_nData);
  a_vals.Create(bound, 2);
  long idx[2] = {0,0};

  int i;
  for (i=0; i<a_nData; i++)
  {
    idx[0] = 0; idx[1] = i;
    a_vals.MultiDimSetAt(idx, CComVariant(a_data[i])); idx[0]++;
    a_vals.MultiDimSetAt(idx, CComVariant(a_startCellId+i)); idx[0]++;
  }
  if (a_stressPeriod > 0)
  {
    idx[0] = 2;
    for (i=0; i<a_nData; i++)
    {
      idx[1] = i;
      a_vals.MultiDimSetAt(idx, CComVariant(a_stressPeriod));
    }
    if (a_etsSegId > 0)
    {
      idx[0] = 3;
      for (i=0; i<a_nData; i++)
      {
        idx[1] = i;
        a_vals.MultiDimSetAt(idx, CComVariant(a_etsSegId));
      }
    }
  }

  return 1;
} // DataToSafeArray::U2DTemplate
// explicit instantiation
int DataToSafeArray::U2DREL (const char *a_table,
                             const char *a_cellField,
                             const char *a_field,
                             int a_nData,
                             int a_startCellId,
                             int a_stressPeriod,
                             int a_etsSegId,
                             const float *a_data,
                             CComSafeArray<VARIANT> &a_vals,
                             CComSafeArray<BSTR> &a_fieldNames,
                             CComBSTR &a_tableName)
{
  return (U2DTemplate(a_table, a_cellField, a_field, a_nData, a_startCellId,
                      a_stressPeriod, a_etsSegId, a_data, a_vals, a_fieldNames,
                      a_tableName));
} // DataToSafeArray::U2DREL
int DataToSafeArray::U2DREL (const char *a_table,
                             const char *a_cellField,
                             const char *a_field,
                             int a_nData,
                             int a_startCellId,
                             int a_stressPeriod,
                             int a_etsSegId,
                             const double *a_data,
                             CComSafeArray<VARIANT> &a_vals,
                             CComSafeArray<BSTR> &a_fieldNames,
                             CComBSTR &a_tableName)
{
  return (U2DTemplate(a_table, a_cellField, a_field, a_nData, a_startCellId,
                      a_stressPeriod, a_etsSegId, a_data, a_vals, a_fieldNames,
                      a_tableName));
} // DataToSafeArray::U2DREL
int DataToSafeArray::U2DINT (const char *a_table,
                             const char *a_cellField,
                             const char *a_field,
                             int a_nData,
                             int a_startCellId,
                             int a_stressPeriod,
                             int a_etsSegId,
                             const int *a_data,
                             CComSafeArray<VARIANT> &a_vals,
                             CComSafeArray<BSTR> &a_fieldNames,
                             CComBSTR &a_tableName)
{
  return (U2DTemplate(a_table, a_cellField, a_field, a_nData, a_startCellId,
                      a_stressPeriod, a_etsSegId, a_data, a_vals, a_fieldNames,
                      a_tableName));
} // DataToSafeArray::U2DINT

///////////////////////////////////////////////////////////////////////////////
// TESTS
///////////////////////////////////////////////////////////////////////////////
#if CXX_TEST
using namespace DataToSafeArray;

#include <DataToSafeArray.t.h>
//------------------------------------------------------------------------------
void DataToSafeArrayT::testOnlyArrayU2DREL ()
{
  CComSafeArray<VARIANT> vals;
  CComSafeArray<BSTR> fields;
  CComBSTR tabName;
  char *table = "EVTArrays";
  char *field = "SURF";
  int nDat(9);
  float d[9] = {1.0f,2.0f,3.0f,4.0f,5.0f,6.0f,7.0f,8.0f,9.0f};
  float *fltPtrNull(0);

  // test nDat < 1 and data = NULL
  TS_ASSERT(!OnlyArrayU2DREL(table, field, 0, d, vals, fields, tabName));
  TS_ASSERT(!OnlyArrayU2DREL(table, field, 0, fltPtrNull, vals, fields,
                             tabName));
  // this should work
  TS_ASSERT(OnlyArrayU2DREL(table, field, nDat, d, vals, fields, tabName));
  TS_ASSERT(tabName == CComBSTR(table));
  TS_ASSERT(fields.GetCount() == 1);
  TS_ASSERT(fields.GetAt(0) == CComBSTR(field));
  CComVariant var;
  long idx[2] = {0,0};
  for (int i=0; i<9; i++)
  {
    idx[1] = i;
    TS_ASSERT(vals.MultiDimGetAt(idx, var) == S_OK)
    TS_ASSERT(var == CComVariant(d[i]));
  }
}
//------------------------------------------------------------------------------
void DataToSafeArrayT::testU2DREL ()
{
  char *table = "EVTArrays";
  char *cField = "IJ";
  char *field = "SURF";
  int nDat(9), startCellId(1), sp(2);
  float d[9] = {1.0f,2.0f,3.0f,4.0f,5.0f,6.0f,7.0f,8.0f,9.0f};
  float *fptrNull(0);
  CComSafeArray<VARIANT> vals;
  CComSafeArray<BSTR> fields;
  CComBSTR tabName;

  // the table and field have already been tested to see if they exist
  // test when nData < 1, startCellId < 1, a_data = NULL
  TS_ASSERT(!U2DREL(table, cField, field, 0, startCellId, sp, -1, d, vals, fields,
                    tabName));
  TS_ASSERT(!U2DREL(table, cField, field, nDat, 0, sp, -1, d, vals, fields,
                    tabName));
  TS_ASSERT(!U2DREL(table, cField, field, nDat, startCellId, sp, -1, fptrNull,
                    vals, fields, tabName));

  // this should work
  TS_ASSERT(U2DREL(table, cField, field, nDat, startCellId, sp, -1, d,
                   vals, fields, tabName));
  TS_ASSERT(tabName == CComBSTR(table));
  TS_ASSERT_EQUALS(vals.GetCount(0), 3);
  TS_ASSERT_EQUALS(vals.GetCount(1), 9);
  TS_ASSERT_EQUALS(fields.GetCount(), 3);
  int i(0);
  TS_ASSERT(fields.GetAt(i++) == CComBSTR(field));
  TS_ASSERT(fields.GetAt(i++) == CComBSTR("IJ"));
  TS_ASSERT(fields.GetAt(i++) == CComBSTR("SPID"));

  long idx[2] = {0,0};
  CComVariant var;
  for (i=0; i<9; i++)
  {
    idx[0] = 0; idx[1] = i;
    TS_ASSERT(vals.MultiDimGetAt(idx, var)==S_OK); idx[0]++;
    TS_ASSERT(var == CComVariant(d[i]));
    TS_ASSERT(vals.MultiDimGetAt(idx, var)==S_OK); idx[0]++;
    TS_ASSERT(var == CComVariant(i+1));
    TS_ASSERT(vals.MultiDimGetAt(idx, var)==S_OK); idx[0]++;
    TS_ASSERT(var == CComVariant(sp));
  }
}

#endif
