// *****************************************************************************
// NOTES:  This code is for editing existing geodatabases, tables, etc.,
//         not creating new ones if they do not already exist.
// *****************************************************************************

//#include "StdAfx.h"
#include "EditGDB.h"
#include "TableDef.h"
#include "CreateGDB.h"

// *****************************************************************************
CEditGDB::CEditGDB (void)
{
  // Register for COM to use the ESRI stuff
  ::CoInitialize(0);

  m_szFile = "";
} // CEditGDB::CEditGDB
// *****************************************************************************
CEditGDB::~CEditGDB (void)
{
  // Unregister COM
  ::CoUninitialize();
} // CEditGDB::~CEditGDB
// *****************************************************************************
// NOTES:  Opens up the geodatabase, without starting to edit
// *****************************************************************************
HRESULT CEditGDB::OpenGeodatabase2 (IWorkspace **ppWorkspace)
{
  bool           bAccess = false;
  bool           bFile = false;
  HRESULT        hr;
  IWorkspacePtr  pTempWorkspace;

  // Create an Access workspace factory
  IWorkspaceFactoryPtr pAccessFactory;
  hr = pAccessFactory.CreateInstance(CLSID_AccessWorkspaceFactory);
  if (FAILED(hr))
    return hr;

  // Create a File database factory
  IWorkspaceFactoryPtr pFileFactory;
  hr = pFileFactory.CreateInstance(CLSID_FileGDBWorkspaceFactory);
  if (FAILED(hr))
    return hr;

  // Try to open the Access workspace
  try {
    hr = pAccessFactory->OpenFromFile(m_szFile, NULL, &pTempWorkspace);
    if (!FAILED(hr))
      bAccess = true;
  }
  catch (...) {
    bAccess = false;
  }

  // Try to open the File workspace
  try {
    if (!bAccess)
      hr = pFileFactory->OpenFromFile(m_szFile, NULL, &pTempWorkspace);
      if (!FAILED(hr))
        bFile = true;
  }
  catch (...) {
    bFile = false;
  }

  if (bAccess || bFile) {
    (*ppWorkspace) = pTempWorkspace;
    if (*ppWorkspace) {
      (*ppWorkspace)->AddRef();
    }
  }
  else {
    return E_INVALIDARG;
  }

  // Make sure we really have a workspace to deal with here
  if (*ppWorkspace == 0)
    return E_FAIL;

  return hr;
} // CEditGDB::OpenGeodatabase2
// *****************************************************************************
// NOTES:  Gets the table of the specified name
// *****************************************************************************
HRESULT CEditGDB::FindTable (IWorkspacePtr ipWorkspace, 
                             CComBSTR a_name, ITable **pTable)
{
  // Check to make sure we've got a workspace
  if (ipWorkspace == 0)
    return S_FALSE;

  ITablePtr        pTempTable;
  IEnumDatasetPtr  pDatasetsEnum;
  IDatasetPtr      pDataset;
  CComBSTR         curname;
  HRESULT          hr;

  // Get a feature workspace, then open the table by name
  IFeatureWorkspacePtr ipFeatureWorkspace(ipWorkspace);

  if (ipFeatureWorkspace != NULL) {
    // Open the table, based on the name
    hr = ipFeatureWorkspace->OpenTable(a_name, &pTempTable);
    if (FAILED(hr))
    {
      CStr tab(a_name);
      CTableDef tableDef;
      // try creating the table
      int id(tableDef.GetTableIdFromString(tab));
      if (id < 0)
        return hr;
      ITablePtr  tTable;
      CCreateGDB c;
      hr = c.AddTable(static_cast<ma_tables>(id), ipWorkspace, &tTable);
      if (FAILED(hr))
        return hr;
      hr = ipFeatureWorkspace->OpenTable(a_name, &pTempTable);
    }
    if (FAILED(hr))
      return hr;

    // If that worked...
    if (pTempTable != NULL) {
      *pTable = pTempTable;
      if (*pTable) {
        (*pTable)->AddRef();
        return S_OK;
      }
      else {
        return S_FALSE;
      }
    }
  }

  return S_FALSE;

  // NOTE:  The code below loops on all of the table datasets in a geodatabase.
  //        It should work, but if you get an HRESULT of -2147220655,
  //        (FDO_E_TABLE_NOT_FOUND), and you can't keep on looping through
  //        the enumeration, it's because ArcView doesn't like tables with
  //        certain names, like "General", "Parameters" or "Top".  In this
  //        case, you need to update the data model so these names are gone.

#if 0
  // Get the datasets list (exclude the list to only be Tables)
  hr = ipWorkspace->get_Datasets(esriDTTable, &pDatasetsEnum);
  if (FAILED(hr))
    return hr;

  // Loop on the datasets
  hr = pDatasetsEnum->Next(&pDataset);
  while (pDataset != NULL) {
    // Get the name of the current dataset
    hr = pDataset->get_Name(&curname);
    if (FAILED(hr))
      return hr;

    // See if it matches
    if (a_name == curname) {
      pTempTable = pDataset;
      *pTable = pTempTable;
    }
    hr = pDatasetsEnum->Next(&pDataset);
  }

  if (*pTable) {
    (*pTable)->AddRef();
    return S_OK;
  }
  else {
    return S_FALSE;
  }
#endif
} // CEditGDB::FindTable
// *****************************************************************************
// NOTES:  Starts editing the workspace
// *****************************************************************************
HRESULT CEditGDB::StartEditing (IWorkspacePtr ipWorkspace)
{
  HRESULT            hr;
  IWorkspaceEditPtr  ipWorkspaceEdit(ipWorkspace);

  // Start editing, disabling the undo/redo logging
  hr = ipWorkspaceEdit->StartEditing(VARIANT_FALSE);

  return(hr);
} // CEditGDB::StartEditing
// *****************************************************************************
// NOTES:  Finishes editing the workspace
// *****************************************************************************
HRESULT CEditGDB::StopEditing (IWorkspacePtr ipWorkspace)
{
  HRESULT            hr;
  VARIANT_BOOL       bEditing;
  IWorkspaceEditPtr  ipWorkspaceEdit(ipWorkspace);

  // Make sure that we were really editing at first
  hr = ipWorkspaceEdit->IsBeingEdited(&bEditing);

  // If editing, stop editing, saving our edits
  if (bEditing == VARIANT_TRUE)
    hr = ipWorkspaceEdit->StopEditing(VARIANT_TRUE);

  return(hr);
} // CEditGDB::StopEditing
// *****************************************************************************
// NOTES:  Add values from a vector of variants into the table passed in.  This
//         is for when you are using a cursor and row buffers to add multiple
//         rows at a time.
//         a_Cursor:  The insert cursor we are writing to
//         a_values:  The values to store in the table (can be mixed type)
//         a_names:  The corresponding table field names the values go in.
//                   Should be the same size as a_values, and only be BSTR's.
// *****************************************************************************
HRESULT CEditGDB::AddValuesToTable2 (ICursorPtr a_Cursor,
                                     ITablePtr a_Table,
                                     CComSafeArray<VARIANT> &a_values,
                                     CComSafeArray<BSTR> &a_names)
{
  HRESULT        hr;
  IFieldsPtr     pFields;
  IFieldPtr      pField;
  IRowBufferPtr  ipRowBuffer;
  long           i, lNumFields;
  unsigned long  j;
  esriFieldType  fType;
  CComBSTR       curfieldname, tempname;
  CComVariant    vOID;

  if (a_Cursor == 0 || a_Table == 0)
    return E_FAIL;

  // Get the fields
  hr = a_Table->get_Fields(&pFields);
  if (FAILED(hr))
    return hr;

  // Get the number of fields
  hr = pFields->get_FieldCount(&lNumFields);
  if (FAILED(hr))
    return hr;

  // Create a new row buffer
  hr = a_Table->CreateRowBuffer(&ipRowBuffer);
  if (FAILED(hr))
    return hr;

  // Loop on all of the fields in the table
  for (i = 0; i < lNumFields; ++i) {
    // Get the field
    hr = pFields->get_Field(i, &pField);
    if (FAILED(hr))
      return hr;

    // See if it's a valid field to write to (not an object ID, shape, etc)
    hr = pField->get_Type(&fType);
    if (FAILED(hr))
      return hr;
    if ((fType == esriFieldTypeOID) ||
        (fType == esriFieldTypeGeometry) ||
        (fType == esriFieldTypeBlob)) // We may have to get rid of this last one
      continue;

    // Now, loop on the field names passed in.  If we find a match, store the
    // associated value on the row
    hr = pField->get_Name(&curfieldname);
    if (FAILED(hr))
      return hr;
    for (j = 0; j < a_names.GetCount(); ++j) {
      tempname = a_names.GetAt(j);
      if (tempname == curfieldname) {
        // We found a match, store the value in the row
        if (j < a_values.GetCount()) {
          hr = ipRowBuffer->put_Value(i, a_values.GetAt(j));
          if (FAILED(hr))
            return hr;
        }
      }
      else if ((curfieldname == "Shead" && tempname == "SHead") ||
               (curfieldname == "Ehead" && tempname == "EHead")) {
        // We found a match, store the value in the row
        // THIS IS A BUGFIX FOR CHD... names defined differently in
        // MF2K project than MODFLOW Data Model
        if (j < a_values.GetCount()) {
          hr = ipRowBuffer->put_Value(i, a_values.GetAt(j));
          if (FAILED(hr))
            return hr;
        }
      }
    }
  }

  // Insert the row in the row buffer
  hr = a_Cursor->InsertRow(ipRowBuffer, &vOID);
  return hr;
} // CEditGDB::AddValuesToTable2 
// *****************************************************************************
// NOTES:  Generic:  Any old Table.
//         The order of the values in the 2D vector should correspond with
//         the field names vector.
//
//         You do not have to write to all of the fields, just don't make
//         a vector of values for them, and don't put the name of the field in
//         the field names vector.
//
// EXAMPLE:  Say you have a table with an Instance, CellID, and Q.
//           You want to make two rows in the table (two values for each field).
//           Make one VEC_VARIANT for each of the fields:
//
//           VEC_VARIANT Instance, CellID, Q;
//
//           Store the values for each row in the vector:
//           Instance.at(0) = 1; Instance.at(1) = 2;
//           CellID.at(0) = 15; CellID.at(1) = 12;
//           Q.at(0) = -44455.0; Q.at(1) = -12345.0;
//
//           Now, make a 2D Variant vector.  Store the previous vectors
//           IN THE SAME ORDER AS THEY APPEAR IN THE TABLE:
//           VEC_VAR2D values;
//           values.push_back(Instance);
//           values.push_back(CellID);
//           values.push_back(Q);
//
//           VEC_BSTR names;
//
//           Store the names for each field in the vector, in correct order:
//           names.at(0)="Instance"; names.at(1)="CellID"; names.at(2)="Q";
//
//           Then call EditTable:
//           EditTable("SomeName", names, values);
// *****************************************************************************
HRESULT CEditGDB::EditTable (CComBSTR a_TableName,
                             CComSafeArray<BSTR> &a_FieldNames,
                             CComSafeArray<VARIANT> &a_2DValues)
{
  int            j;
  long           numrows(0);
  unsigned long  i, numcols(0);
  HRESULT        hr;
  ITablePtr      ipTable;
  IWorkspacePtr  ipWorkspace;
  ICursorPtr     ipCursor;

  // Open the Geodatabase
  hr = OpenGeodatabase2(&ipWorkspace);
  if (FAILED(hr))
    return hr;

  // Find the table
  hr = FindTable(ipWorkspace, a_TableName, &ipTable);
  if (FAILED(hr))
    return hr;

  // Create an insert cursor for the table
  hr = ipTable->Insert(VARIANT_TRUE, &ipCursor);
  if (FAILED(hr))
    return hr;

  // Get the number of rows we will be creating
  long idx[2];
  VARIANT v;
  numrows = a_FieldNames.GetCount();
  UINT uDim(1);
  numcols = a_2DValues.GetCount(uDim);

  for (i = 0; i < numcols; ++i) {
    CComSafeArray<VARIANT> values;
    idx[1] = i;
    for (j = 0; j < numrows; ++j)
    {
      idx[0] = j;
      a_2DValues.MultiDimGetAt(idx, v);
      values.Add(v);
    }
    hr = AddValuesToTable2(ipCursor, ipTable, values, a_FieldNames);
    if (FAILED(hr))
      return hr;
  }
  hr = ipCursor->Flush();
  return hr;
} // CEditGDB::EditTable
// *****************************************************************************
// NOTES:  Generic:  Any old Table.
//         The order of the values in the 2D vector should correspond with
//         the field names vector.
//
//         Used when you already have a table, but simply need to edit the rows.
//         For example, you may have written rows in a table, but only for
//         a single field.  Now, you need to go back and fill in other fields.
//         Data should be in the same order.
//
//         Simply omit the field you already wrote.
//
//         If you want to do a query and only update certain rows, add an
//         SQL statement to the a_WhereClause parameter, otherwise leave it
//         empty.  An example of a WhereClause might be "SPID = 1".  Strings
//         (not field names) must be in single quotes.
//
// EXAMPLE:  Say you have a table with an Instance, CellID, and Q.
//           You want to make two rows in the table (two values for each field).
//           Make one VEC_VARIANT for each of the fields:
//
//           VEC_VARIANT Instance, CellID, Q;
//
//           Store the values for each row in the vector:
//           Instance.at(0) = 1; Instance.at(1) = 2;
//           CellID.at(0) = 15; CellID.at(1) = 12;
//           Q.at(0) = -44455.0; Q.at(1) = -12345.0;
//
//           Now, make a 2D Variant vector.  Store the previous vectors
//           IN THE SAME ORDER AS THEY APPEAR IN THE TABLE:
//           VEC_VAR2D values;
//           values.push_back(Instance);
//           values.push_back(CellID);
//           values.push_back(Q);
//
//           VEC_BSTR names;
//
//           Store the names for each field in the vector, in correct order:
//           names.at(0)="Instance"; names.at(1)="CellID"; names.at(2)="Q";
//
//           Then call EditTable:
//           EditTable("SomeName", names, values);
// *****************************************************************************
HRESULT CEditGDB::EditTableExistingRows (CComBSTR a_TableName,
                                         CComSafeArray<BSTR> &a_UsedFieldNames,
                                         CComSafeArray<VARIANT> &a_2DValues,
                                         int &a_nNumRowsEdited,
                                         CComBSTR a_WhereClause /* = "" */)
{
  int              i, j;
  long             numrows(0), numcols(0), oid, minOID;
  HRESULT          hr;
  ITablePtr        ipTable;
  IWorkspacePtr    ipWorkspace;
  ICursorPtr       ipCursor;
  IRowPtr          ipRow;
  IQueryFilterPtr  ipQueryFilter = 0;

  a_nNumRowsEdited = 0;

  // Open the Geodatabase
  hr = OpenGeodatabase2(&ipWorkspace);
  if (FAILED(hr))
    return hr;

  // Find the table
  hr = FindTable(ipWorkspace, a_TableName, &ipTable);
  if (FAILED(hr))
    return hr;

  // Create an update cursor for the table to get all of the rows that will
  // allow us to update rows.  Filter if the WhereClause is passed in.
  if (a_WhereClause.Length() > 0) {
    hr = ipQueryFilter.CreateInstance(CLSID_QueryFilter);
    if (FAILED(hr)) return hr;

    hr = ipQueryFilter->put_WhereClause(a_WhereClause);
    if (FAILED(hr)) return hr;
  }
  hr = ipTable->Update(ipQueryFilter, true, &ipCursor);
  if (FAILED(hr))
    return hr;

  // Get the number of rows we will be creating
  numrows = a_UsedFieldNames.GetCount();
  numcols = a_2DValues.GetCount(1);

  // get the min OID
  for (i = 0; i < numcols; ++i) {
    // Get the row we want to write to
    hr = ipCursor->NextRow(&ipRow);
    if (FAILED(hr) || 0 == ipRow) return hr;
    ipRow->get_OID(&oid);
    if (i==0 || oid < minOID)
      minOID = oid;
  }

  hr = ipTable->Update(ipQueryFilter, true, &ipCursor);
  if (FAILED(hr))
    return hr;
  // Start at the first row in the update cursor, and keep going as long
  // as we have data in the vectors or a new row to write to
  for (i = 0; i < numcols; ++i) {
    // Get the row we want to write to
    hr = ipCursor->NextRow(&ipRow);
    if (FAILED(hr) || 0 == ipRow) return hr;
    ipRow->get_OID(&oid);

    CComSafeArray<VARIANT> values;
    long idx[2];
    VARIANT v;
    for (j = 0; j < numrows; ++j)
    {
      idx[0] = j;
      idx[1] = oid-minOID;
      if (i != (oid-minOID))
        i = i;
      a_2DValues.MultiDimGetAt(idx, v);
      values.Add(v);
    }
    hr = AppendValuesToTable(ipRow, ipTable, values, a_UsedFieldNames);
    if (FAILED(hr))
      return hr;

    // Update the row -- stores the changes we made
    hr = ipCursor->UpdateRow(ipRow);
    if (FAILED(hr))
      return hr;

    ++a_nNumRowsEdited;
  }
  hr = ipCursor->Flush();
  return hr;
} // CEditGDB::EditTableExistingRows
// *****************************************************************************
// NOTES:  Add values from a vector of variants into the table passed in.  This
//         is for when you are using a cursor to add additional data to an
//         existing row.
//         a_Row:  The row we are writing to
//         a_values:  The values to store in the table (can be mixed type)
//         a_names:  The corresponding table field names the values go in.
//                   Should be the same size as a_values, and only be BSTR's.
// *****************************************************************************
HRESULT CEditGDB::AppendValuesToTable (IRowPtr a_Row,
                                       ITablePtr a_Table,
                                       CComSafeArray<VARIANT> &a_values,
                                       CComSafeArray<BSTR> &a_usednames)
{
  HRESULT        hr;
  IFieldsPtr     pFields;
  IFieldPtr      pField;
  long           i, lNumFields;
  unsigned long  j;
  esriFieldType  fType;
  CComBSTR       curfieldname, tempname;
  CComVariant    vOID;

  if (a_Row == 0 || a_Table == 0)
    return E_FAIL;

  // Get the fields
  hr = a_Table->get_Fields(&pFields);
  if (FAILED(hr))
    return hr;

  // Get the number of fields
  hr = pFields->get_FieldCount(&lNumFields);
  if (FAILED(hr))
    return hr;

  // Loop on all of the fields in the table
  for (i = 0; i < lNumFields; ++i) {
    // Get the field
    hr = pFields->get_Field(i, &pField);
    if (FAILED(hr))
      return hr;

    // See if it's a valid field to write to (not an object ID, shape, etc)
    hr = pField->get_Type(&fType);
    if (FAILED(hr))
      return hr;
    if ((fType == esriFieldTypeOID) ||
        (fType == esriFieldTypeGeometry) ||
        (fType == esriFieldTypeBlob)) // We may have to get rid of this last one
      continue;

    // Now, loop on the field names passed in.  If we find a match, store the
    // associated value on the row
    hr = pField->get_Name(&curfieldname);
    if (FAILED(hr))
      return hr;
    for (j = 0; j < a_usednames.GetCount(); ++j) {
      tempname = a_usednames.GetAt(j);
      if (tempname == curfieldname) {
        // We found a match, store the value in the row
        if (j < a_values.GetCount()) {
          hr = a_Row->put_Value(i, a_values.GetAt(j));
          if (FAILED(hr))
            return hr;
        }
      }
    }
  }

  return hr;
} // CEditGDB::AppendValuesToTable
// *****************************************************************************
// NOTES:  Loops through the tables in the workspace, and tries to find the
//         table with the name passed in
// *****************************************************************************
bool CEditGDB::TableExists (IFeatureWorkspacePtr ipFW, CComBSTR szName)
{
  HRESULT hr;
  IWorkspacePtr ipWS(ipFW);
  IEnumDatasetNamePtr ipEnumDatasetName;
  IDatasetNamePtr ipDatasetName;
  CComBSTR bstrName;

  if (ipWS)
  {
    hr = ipWS->get_DatasetNames(esriDTTable, &ipEnumDatasetName);
    if (FAILED(hr)) return false;

    if (ipEnumDatasetName)
    {
      hr = ipEnumDatasetName->Next(&ipDatasetName);
      if (FAILED(hr)) return false;

      while (ipDatasetName)
      {
        hr = ipDatasetName->get_Name(&bstrName);
        if (FAILED(hr)) return false;

        if (bstrName == szName)
          return true;

        hr = ipEnumDatasetName->Next(&ipDatasetName);
        if (FAILED(hr)) return false;
      }
    }
  }

  return false;
} // CEditGDB::TableExists
// *****************************************************************************
// NOTES:  Creates missing tables for packages where only part of the tables
//         were created.  This could occur when parameters are used for
//         a package like LPF, where none of the arrays get written and so
//         that the LPFProperties table never gets created.  Some of the
//         AHGW tools need that table, even if it's empty, to function.
// *****************************************************************************
HRESULT CEditGDB::AddMissingTables ()
{
  HRESULT        hr;
  ITablePtr      ipTable;
  IWorkspacePtr  ipWorkspace;
  CComBSTR       szMainTable;
  CCreateGDB     create;

  // Open the Geodatabase
  hr = OpenGeodatabase2(&ipWorkspace);
  if (FAILED(hr))
    return hr;

  // Set up the class for adding tables
  create.SetGDBFile2(m_szFile);

  // Handle the RCH Package
  // Find the main table:  RCHVars
  szMainTable = CComBSTR("RCHVars");
  if (TableExists(ipWorkspace, szMainTable))
  {
    // Check the RCHArrays table
    szMainTable = CComBSTR("RCHArrays");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(RCHARRAYS_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }

    // Check the RCHArrayMult table
    szMainTable = CComBSTR("RCHArrayMult");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(RCHARRAYMULT_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }
  }

  // Handle the EVT Package
  // Find the main table:  EVTVars
  szMainTable = CComBSTR("EVTVars");
  if (TableExists(ipWorkspace, szMainTable))
  {
    // Check the EVTArrays table
    szMainTable = CComBSTR("EVTArrays");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(EVTARRAYS_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }

    // Check the EVTArrayMult table
    szMainTable = CComBSTR("EVTArrayMult");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(EVTARRAYMULT_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }
  }

  // Handle the Basic Package
  // Find the main table:  BASVars
  szMainTable = CComBSTR("BASVars");
  if (TableExists(ipWorkspace, szMainTable))
  {
    // Check the Basic table
    szMainTable = CComBSTR("Basic");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(BASIC_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }

    // Check the BasicArrayMult table
    szMainTable = CComBSTR("BasicArrayMult");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(BASIC_MULT_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }
  }

  // Handle the BCF6 Package
  // Find the main table:  BCFVars
  szMainTable = CComBSTR("BCFVars");
  if (TableExists(ipWorkspace, szMainTable))
  {
    // Check the BCFLayers table
    szMainTable = CComBSTR("BCFLayers");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(BCFLAYERS_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }

    // Check the BCFProperties table
    szMainTable = CComBSTR("BCFProperties");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(BCFPROPERTIES_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }
  }

  // Handle the LPF Package
  // Find the main table:  LPFVars
  szMainTable = CComBSTR("LPFVars");
  if (TableExists(ipWorkspace, szMainTable))
  {
    // Check the LPFLayers table
    szMainTable = CComBSTR("LPFLayers");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(LPFLAYERS_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }

    // Check the LPFProperties table
    szMainTable = CComBSTR("LPFProperties");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(LPFPROPERTIES_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }
  }

  // Handle the HUF Package
  // Find the main table:  HUFVars
  szMainTable = CComBSTR("HUFVars");
  if (TableExists(ipWorkspace, szMainTable))
  {
    // Check the HUFLayers table
    szMainTable = CComBSTR("HUFLayers");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(HUFLAYERS_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }

    // Check the HUFWetDry table
    szMainTable = CComBSTR("HUFWetDry");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(HUFWETDRY_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }

    // Check the HUFUnits table
    szMainTable = CComBSTR("HUFUnits");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(HUFUNITS_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }

    // Check the HUFUnitProps table
    szMainTable = CComBSTR("HUFUnitProps");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(HUFUNITPROPS_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }
  }

  // Handle the ETS Package
  // Find the main table:  ETSVars
  szMainTable = CComBSTR("ETSVars");
  if (TableExists(ipWorkspace, szMainTable))
  {
    // Check the ETSArrays table
    szMainTable = CComBSTR("ETSArrays");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(ETSARRAYS_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }

    // Check the ETSArrayMult table
    szMainTable = CComBSTR("ETSArrayMult");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(ETSARRAYMULT_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }

    // Check the ETSSegArrays table
    szMainTable = CComBSTR("ETSSegArrays");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(ETSSEGARRAYS_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }

    // Check the ETSSegArrayMult table
    szMainTable = CComBSTR("ETSSegArrayMult");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(ETSSEGARRAYMULT_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }
  }

  // Handle the Stream Package
  // Find the main table:  STRVars
  szMainTable = CComBSTR("STRVars");
  if (TableExists(ipWorkspace, szMainTable))
  {
    // Check the STRItrib table
    szMainTable = CComBSTR("STRItrib");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(STRITRIB_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }

    // Check the STRIupseg table
    szMainTable = CComBSTR("STRIupseg");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(STRIUPSEG_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }

    // Check the STRReaches table
    szMainTable = CComBSTR("STRReaches");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(STRREACHES_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }
  }

  // Handle the SFR Package
  // Find the main table:  SFRVars
  szMainTable = CComBSTR("SFRVars");
  if (TableExists(ipWorkspace, szMainTable))
  {
    // Check the SFRReaches table
    szMainTable = CComBSTR("SFRReaches");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(SFRREACHES_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }

    // Check the SFRFlowTab table
    szMainTable = CComBSTR("SFRFlowTab");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(SFRFLOWTAB_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }

    // Check the SFRSegments table
    szMainTable = CComBSTR("SFRSegments");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(SFRSEGMENTS_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }

    // Check the SFRXsect table
    szMainTable = CComBSTR("SFRXsect");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(SFRXSECT_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }
  }

  // Handle the LAK Package
  // Find the main table:  LAKVars
  szMainTable = CComBSTR("LAKVars");
  if (TableExists(ipWorkspace, szMainTable))
  {
    // Check the LAKArrays table
    szMainTable = CComBSTR("LAKArrays");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(LAKARRAYS_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }

    // Check the LAKStages table
    szMainTable = CComBSTR("LAKStages");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(LAKSTAGES_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }

    // Check the LAKSublakes table
    szMainTable = CComBSTR("LAKSublakes");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(LAKSUBLAKES_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }

    // Check the LAKRates table
    szMainTable = CComBSTR("LAKRates");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(LAKRATES_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }
  }

  // Handle the UZF Package
  // Find the main table:  UZFVars
  szMainTable = CComBSTR("UZFVars");
  if (TableExists(ipWorkspace, szMainTable))
  {
    // Check the UZFArrays table
    szMainTable = CComBSTR("UZFArrays");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(UZFARRAYS_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }

    // Check the UZFArrayMult table
    szMainTable = CComBSTR("UZFArrayMult");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(UZFARRAYMULT_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }

    // Check the UZFGages table
    szMainTable = CComBSTR("UZFGages");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(UZFGAGES_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }

    // Check the UZFStressArrays table
    szMainTable = CComBSTR("UZFStressArrays");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(UZFSTRESSARRAYS_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }

    // Check the UZFStressArrayMult table
    szMainTable = CComBSTR("UZFStressArrayMult");
    if (!TableExists(ipWorkspace, szMainTable))
    {
      hr = create.AddTable(UZFSTRESSARRAYMULT_TABLE, ipWorkspace, &ipTable);
      if (FAILED(hr)) return hr;
    }
  }

  return hr;
} // CEditGDB::AddMissingTables
// *****************************************************************************
// NOTES:  Deletes all rows in the table
// *****************************************************************************
HRESULT CEditGDB::DeleteTable (CComBSTR a_TableName)
{
  HRESULT        hr;
  ITablePtr      ipTable;
  IWorkspacePtr  ipWorkspace;
  IDatasetPtr    ipDataset;

  // Open the Geodatabase
  hr = OpenGeodatabase2(&ipWorkspace);
  if (FAILED(hr))
    return hr;

  // Find the table
  hr = FindTable(ipWorkspace, a_TableName, &ipTable);
  if (FAILED(hr))
    return hr;

  // delete the table
  ipDataset = ipTable;
  if (ipDataset)
  {
    VARIANT_BOOL bDelete;
    hr = ipDataset->CanDelete(&bDelete);
    if (FAILED(hr)) return hr;
    if (bDelete == VARIANT_TRUE)
    {
      hr = ipDataset->Delete();
      if (FAILED(hr)) return hr;
    }
  }

  return S_OK;
} // CEditGDB::DeleteAllRowsInTable

