// *****************************************************************************
// TableDef.cpp
//
// Holds the field names and field types for all of the tables in the
// Modflow Data model.  Whenever changes are made to the data model, make
// any and all changes here as well.
//
// Vectors should be stored so that the field names and field types match up
// in the same order.
// *****************************************************************************

//#include "stdafx.h"

#include "TableDef.h"

#include "MDMDomains.h"

#include <map>
#include <set>

// *****************************************************************************
// NOTES:  Internal functions
// *****************************************************************************
namespace
{
  static std::map<CStr, std::set<CStr> >& iGetAuxCols ()
  {
    static std::map<CStr, std::set<CStr> > m_map;
    return m_map;
  } // iGetAuxCols
}
// *****************************************************************************
// NOTES:  Add AUX columns to the table definition
// *****************************************************************************
void CTableDef::AddAuxColumnToTableDef (CStr tablename,
                                        CStr colName)
{
  std::map<CStr, std::set<CStr> >& m(iGetAuxCols());
  m[tablename].insert(colName);
} // CTableDef::AddAuxColumnToTableDef
// *****************************************************************************
// NOTES:  Gets the field names and ESRI field type for a particular table
// *****************************************************************************
int CTableDef::GetData (int TableID,
                        CComSafeArray<BSTR> &vN,
                        VEC_FTYPE &vT,
                        VEC_DOMAIN& vD,
                        CComBSTR &name)
{
  int rval = 1;

  // Clear out the incoming vectors
  vN.Destroy();
  vT.clear();

  switch (TableID) {
    // =========================================================================
    // Global Settings
    case MDFGLOBALS_TABLE:
      name = "MDFGlobals";
      vN.Add(CComBSTR("MODFLOW_Version"));  vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("DataModelVersion")); vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("NameFilePath"));     vT.push_back(esriFieldTypeString);
      break;

    // =========================================================================
    // Name file
    case NAMEFILE_TABLE:
      name = "NameFile";
      vN.Add(CComBSTR("FileType"));        vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("Nunit"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Fname"));           vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("Use"));             vT.push_back(esriFieldTypeSmallInteger);
      break;

    // =========================================================================
    // Discretization (DIS)
    case DELRC_TABLE:
      name = "DELRC";
      vN.Add(CComBSTR("Direction"));       vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("Num"));             vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Width"));           vT.push_back(esriFieldTypeDouble);
      vD.push_back(CMDMDomains::Instance().GetDirectionDomain());
      vD.push_back(NULL);
      vD.push_back(NULL);
      break;
    case STRESSPERIODS_TABLE:
      name = "StressPeriods";
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("PERLEN"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("NSTP"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("TSMULT"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("SSorTr"));          vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("RIV_ITMP"));        vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("WEL_ITMP"));        vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("DRN_ITMP"));        vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("GHB_ITMP"));        vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("CHD_ITMP"));        vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("RCH_INRECH"));      vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("RCH_INIRCH"));      vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("EVT_INSURF"));      vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("EVT_INEVTR"));      vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("EVT_INEXDP"));      vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("EVT_INIEVT"));      vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("MNW1_ITMP"));       vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("STR_ITMP"));        vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("STR_IRDFLG"));      vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("STR_IPTFLG"));      vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("SFR_ITMP"));        vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("SFR_IRDFLG"));      vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("SFR_IPTFLG"));      vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("LAK_ITMP"));        vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("LAK_ITMP1"));       vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("LAK_LWRT"));        vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("DRT_ITMP"));        vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("ETS_INETSS"));      vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("ETS_INETSR"));      vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("ETS_INETSX"));      vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("ETS_INIETS"));      vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("ETS_INSGDF"));      vT.push_back(esriFieldTypeSmallInteger);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(CMDMDomains::Instance().GetSteadyStateDomain());
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      break;
    case DISLAYERS_TABLE:
      name = "DISLayers";
      vN.Add(CComBSTR("Layer"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("LAYCBD"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("AM_TopElev"));      vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_BotmElev"));     vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_BotmElevCBD"));  vT.push_back(esriFieldTypeDouble);
      break;
    case TOPELEV_TABLE:
      name = "TopElev";
      vN.Add(CComBSTR("IJK"));             vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("TopElev"));         vT.push_back(esriFieldTypeDouble);
      break;
    case BOTMELEV_TABLE:
      name = "BotmElev";
      vN.Add(CComBSTR("IJK"));             vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("BotmElev"));        vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("BotmElevCBD"));     vT.push_back(esriFieldTypeDouble);
      break;
    case DISVARS_TABLE:
      name = "DISVars";
      vN.Add(CComBSTR("NLAY"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("NROW"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("NCOL"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("ITMUNI"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("LENUNI"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("RefTime"));         vT.push_back(esriFieldTypeDate);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(CMDMDomains::Instance().GetTimeUnitDomain());
      vD.push_back(CMDMDomains::Instance().GetLengthUnitDomain());
      vD.push_back(NULL);
      break;

    // =========================================================================
    // Basic (BAS6)
    case BASICVARS_TABLE:
      name = "BASVars";
      vN.Add(CComBSTR("Options"));         vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("HNOFLO"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("HEADNG1"));         vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("HEADNG2"));         vT.push_back(esriFieldTypeString);
      break;
    case BASIC_TABLE:
      name = "Basic";
      vN.Add(CComBSTR("IJK"));             vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("IBOUND"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("STRT"));            vT.push_back(esriFieldTypeDouble);
      break;
    case BASIC_MULT_TABLE:
      name = "BasicArrayMult";
      vN.Add(CComBSTR("Layer"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("AM_IBOUND"));       vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("AM_STRT"));         vT.push_back(esriFieldTypeDouble);
      break;

    // =========================================================================
    // Output Control
    case OCVARS_TABLE:
      name = "OCVars";
      vN.Add(CComBSTR("IHEDFM"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("IDDNFM"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("IHEDUN"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("IDDNUN"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("IBDOPT"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("IAUXSV"));          vT.push_back(esriFieldTypeSmallInteger);
      vD.push_back(CMDMDomains::Instance().GetOCVarsFileFormatDomain());
      vD.push_back(CMDMDomains::Instance().GetOCVarsFileFormatDomain());
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      break;
    case OCTS_TABLE:
      name = "OCTS";
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("TSNum"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("IHDDFL"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("IBUDFL"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("ICBCFL"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("Hdpr"));            vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("Ddpr"));            vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("Hdsv"));            vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("Ddsv"));            vT.push_back(esriFieldTypeSmallInteger);
      break;
    case CBFLAGS_TABLE:
      name = "CBFlags";
      vN.Add(CComBSTR("IBCFCB"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("ILPFCB"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("IRIVCB"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("IWELCB"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("IDRNCB"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("IGHBCB"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("IRCHCB"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("IEVTCB"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("IMNW1CB"));         vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("ILKCB"));           vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("IHUFCB"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("ISTCB1_STR"));      vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("ISTCB2_STR"));      vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("ISTCB1_SFR"));      vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("ISTCB2_SFR"));      vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("IDRTCB"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("IETSCB"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("IUZFCB1"));         vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("IUZFCB2"));         vT.push_back(esriFieldTypeSmallInteger);
      break;

    // =========================================================================
    // Block-Centered Flow (BCF6)
    case BCFVARS_TABLE:
      name = "BCFVars";
      vN.Add(CComBSTR("HDRY"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("IWDFLG"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("WETFCT"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("IWETIT"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("IHDWET"));          vT.push_back(esriFieldTypeSmallInteger);
      break;
    case BCFLAYERS_TABLE:
      name = "BCFLayers";
      vN.Add(CComBSTR("Layer"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("LAYAVG"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("LAYCON"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("TRPY"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_Sf1"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_Tran"));         vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_HY"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_Vcont"));        vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_Sf2"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_WETDRY"));       vT.push_back(esriFieldTypeDouble);
      vD.push_back(NULL);
      vD.push_back(CMDMDomains::Instance().GetBCFMeanDomain());
      vD.push_back(CMDMDomains::Instance().GetBCFLayerTypeDomain());
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      break;
    case BCFPROPERTIES_TABLE:
      name = "BCFProperties";
      vN.Add(CComBSTR("IJK"));             vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("HY"));              vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Tran"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Vcont"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Sf1"));             vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Sf2"));             vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("WETDRY"));          vT.push_back(esriFieldTypeDouble);
      break;

    // =========================================================================
    // Layer-Property Flow (LPF)
    case LPFVARS_TABLE:
      name = "LPFVars";
      vN.Add(CComBSTR("HDRY"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("WETFCT"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("IWETIT"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("IHDWET"));          vT.push_back(esriFieldTypeSmallInteger);
      break;
    case LPFLAYERS_TABLE:
      name = "LPFLayers";
      vN.Add(CComBSTR("Layer"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("LAYTYP"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("LAYAVG"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("CHANI"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("LAYVKA"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("LAYWET"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("AM_HK"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_HANI"));         vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_VKA"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_Ss"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_Sy"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_WETDRY"));       vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_VKCB"));         vT.push_back(esriFieldTypeDouble);
      vD.push_back(NULL);
      vD.push_back(CMDMDomains::Instance().GetLPFLayerTypeDomain());
      vD.push_back(CMDMDomains::Instance().GetLPFMeanDomain());
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      break;
    case LPFPROPERTIES_TABLE:
      name = "LPFProperties";
      vN.Add(CComBSTR("IJK"));             vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("HK"));              vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("HANI"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("VKA"));             vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Ss"));              vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Sy"));              vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("WETDRY"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("VKCB"));            vT.push_back(esriFieldTypeDouble);
      break;

    // =========================================================================
    // List-Based Stress Packages
    case RIV_TABLE:
      name = "RIV";
      vN.Add(CComBSTR("IJK"));             vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Stage"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Cond"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Rbot"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("IFACE"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Condfact"));        vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("SourceID"));        vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("LinearScale"));     vT.push_back(esriFieldTypeDouble);
      break;
    case CHD_TABLE:
      name = "CHD";
      vN.Add(CComBSTR("IJK"));             vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Shead"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Ehead"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Shdfact"));         vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Ehdfact"));         vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("SourceID"));        vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("LinearScale"));     vT.push_back(esriFieldTypeDouble);
      break;
    case WEL_TABLE:
      name = "WEL";
      vN.Add(CComBSTR("IJK"));             vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Q"));               vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Qfact"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("IFACE"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SourceID"));        vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("LinearScale"));     vT.push_back(esriFieldTypeDouble);
      break;
    case DRN_TABLE:
      name = "DRN";
      vN.Add(CComBSTR("IJK"));             vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Elevation"));       vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Cond"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("IFACE"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Condfact"));        vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("SourceID"));        vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("LinearScale"));     vT.push_back(esriFieldTypeDouble);
      break;
    case GHB_TABLE:
      name = "GHB";
      vN.Add(CComBSTR("IJK"));             vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("BHead"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Cond"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("IFACE"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Condfact"));        vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("SourceID"));        vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("LinearScale"));     vT.push_back(esriFieldTypeDouble);
      break;
    case HFB6_TABLE:
      name = "HFB6";
      vN.Add(CComBSTR("IJK1"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("IJK2"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Hydchr"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Factor"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("SourceID"));        vT.push_back(esriFieldTypeInteger);
      break;
    case DRT_TABLE:
      name = "DRT";
      vN.Add(CComBSTR("IJK"));             vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Elevation"));       vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Cond"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("IFACE"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Condfact"));        vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("LayR"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("RowR"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("ColR"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Rfprop"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("SourceID"));        vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("LinearScale"));     vT.push_back(esriFieldTypeDouble);
      break;

    // =========================================================================
    // Multi-node Well (MNW)
    case MNW1_TABLE:
      name = "MNW1";
      vN.Add(CComBSTR("IJK"));             vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("WellID"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Qdes"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("QwVal"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Rw"));              vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Skin"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Hlim"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Href"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("DD"));              vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Iwgrp"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Cp_C"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("QCUT"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Qfrcmn"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Qfrcmx"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("SITE"));            vT.push_back(esriFieldTypeString);
      break;
    case MNW1VARS_TABLE:
      name = "MNW1Vars";
      vN.Add(CComBSTR("IWELPT"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("NOMOITER"));        vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("kspref"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("LOSSTYPE"));        vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("PLossMNW"));        vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("PREFIX"));          vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("WEL1_FILE"));       vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("iunw1"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("BYNODE_FILE"));     vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("iunby"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("BYNODE_ALLTIME"));  vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("QSUM_FILE"));       vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("iunqs"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("QSUM_ALLTIME"));    vT.push_back(esriFieldTypeSmallInteger);
      break;

    // =========================================================================
    // Recharge (RCH)
    case RCHVARS_TABLE:
      name = "RCHVars";
      vN.Add(CComBSTR("NRCHOP"));          vT.push_back(esriFieldTypeSmallInteger);
      vD.push_back(CMDMDomains::Instance().GetRechargeOptionDomain());
      break;
    case RCHARRAYMULT_TABLE:
      name = "RCHArrayMult";
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("AM_RECH"));         vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_IRCH"));         vT.push_back(esriFieldTypeInteger);
      break;
    case RCHARRAYS_TABLE:
      name = "RCHArrays";
      vN.Add(CComBSTR("IJ"));              vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("RECH"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("IRCH"));            vT.push_back(esriFieldTypeInteger);
      break;

    // =========================================================================
    // Evapotranspiration (EVT)
    case EVTVARS_TABLE:
      name = "EVTVars";
      vN.Add(CComBSTR("NEVTOP"));          vT.push_back(esriFieldTypeSmallInteger);
      vD.push_back(CMDMDomains::Instance().GetEvapotranspirationOptionDomain());
      break;
    case EVTARRAYMULT_TABLE:
      name = "EVTArrayMult";
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("AM_SURF"));         vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_EVTR"));         vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_EXPD"));         vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_IEVT"));         vT.push_back(esriFieldTypeInteger);
      break;
    case EVTARRAYS_TABLE:
      name = "EVTArrays";
      vN.Add(CComBSTR("IJ"));              vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SURF"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("EVTR"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("EXPD"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("IEVT"));            vT.push_back(esriFieldTypeInteger);
      break;

    // =========================================================================
    // Evapotranspiration Segments (ETS)
    case ETSVARS_TABLE:
      name = "ETSVars";
      vN.Add(CComBSTR("NETSOP"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("NETSEG"));          vT.push_back(esriFieldTypeInteger);
      break;

    case ETSARRAYS_TABLE:
      name = "ETSArrays";
      vN.Add(CComBSTR("IJ"));              vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("ETSS"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("ETSR"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("ETSX"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("IETS"));            vT.push_back(esriFieldTypeInteger);
      break;

    case ETSARRAYMULT_TABLE:
      name = "ETSArrayMult";
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("AM_ETSS"));         vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_ETSR"));         vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_ETSX"));         vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_IETS"));         vT.push_back(esriFieldTypeInteger);
      break;

    case ETSSEGARRAYS_TABLE:
      name = "ETSSegArrays";
      vN.Add(CComBSTR("IJ"));              vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SEGID"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("PXDP"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("PETM"));            vT.push_back(esriFieldTypeDouble);
      break;

    case ETSSEGARRAYMULT_TABLE:
      name = "ETSSegArrayMult";
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SEGID"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("AM_PXDP"));         vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_PETM"));         vT.push_back(esriFieldTypeDouble);
      break;

    // =========================================================================
    // Solvers
    case SIP_TABLE:
      name = "SIP";
      vN.Add(CComBSTR("MXITER"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("NPARM"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("ACCL"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("HCLOSE"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("IPCALC"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("WSEED"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("IPRSIP"));          vT.push_back(esriFieldTypeInteger);
      break;
    case PCG_TABLE:
      name = "PCG";
      vN.Add(CComBSTR("MXITER"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("ITER1"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("NPCOND"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("HCLOSE"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("RCLOSE"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("RELAX"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("NBPOL"));           vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("IPRPCG"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("MUTPCG"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("DAMP"));            vT.push_back(esriFieldTypeDouble);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(CMDMDomains::Instance().GetPCGConvergenceDomain());
      vD.push_back(NULL);
      break;
    case GMG_TABLE:
      name = "GMG";
      vN.Add(CComBSTR("RCLOSE"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("IITER"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("HCLOSE"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("MXITER"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("DAMP"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("IADAMP"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("IOUTGMG"));         vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("ISM"));             vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("ISC"));             vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("RELAX"));           vT.push_back(esriFieldTypeDouble);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(CMDMDomains::Instance().GetGMGOutputDomain());
      vD.push_back(NULL);
      vD.push_back(CMDMDomains::Instance().GetGMGSemicoarseningDomain());
      vD.push_back(NULL);
      break;
    case DE4_TABLE:
      name = "DE4";
      vN.Add(CComBSTR("ITMX"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("MXUP"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("MXLOW"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("MXBW"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("IFREQ"));           vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("MUTD4"));           vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("ACCL"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("HCLOSE"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("IPRD4"));           vT.push_back(esriFieldTypeInteger);
      break;
    case LMG_TABLE:
      name = "LMG";
      vN.Add(CComBSTR("STOR1"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("STOR2"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("STOR3"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("ICG"));             vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("MXITER"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("MXCYC"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("BCLOSE"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("DAMP"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("IOUTAMG"));         vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("DUP"));             vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("DLOW"));            vT.push_back(esriFieldTypeDouble);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(CMDMDomains::Instance().GetLMGPrintDomain());
      vD.push_back(NULL);
      vD.push_back(NULL);
      break;
    case SOR_TABLE:
      name = "SOR";
      vN.Add(CComBSTR("MXITER"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("ACCL"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("HCLOSE"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("IPSOR"));           vT.push_back(esriFieldTypeInteger);
      break;

    // =========================================================================
    // Parameters
    case PARAMS_TABLE:
      name = "Params";
      vN.Add(CComBSTR("ParamID"));         vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("PARNAM"));          vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("PARTYPE"));         vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("TableCode"));       vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("FieldCode"));       vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("Parval"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Keyval"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("B"));               vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("BL"));              vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("BU"));              vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("BSCAL"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("LN"));              vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("ISENS"));           vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("LogInterp"));       vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("LogMinVal"));       vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Tied"));            vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("Source"));          vT.push_back(esriFieldTypeSmallInteger);
      break;
    case CLUSTERS_TABLE:
      name = "Clusters";
      vN.Add(CComBSTR("ClusterID"));       vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("ParInstID"));       vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Layer"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("MultID"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("ZoneID"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("HGUNAME"));         vT.push_back(esriFieldTypeString);
      break;
    case IZ_TABLE:
      name = "IZ";
      vN.Add(CComBSTR("ClusterID"));       vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("IZ"));              vT.push_back(esriFieldTypeInteger);
      break;
    case PILOTPTS_TABLE:
      name = "PilotPts";
      vN.Add(CComBSTR("ParamID"));         vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("PointName"));       vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("Parval"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("B"));               vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("ISENS"));           vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("MultID"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SourceID"));        vT.push_back(esriFieldTypeInteger);
      break;
    case MULTIPLIERS_TABLE:
      name = "Multipliers";
      vN.Add(CComBSTR("MultID"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("IJ"));              vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("RMLT"));            vT.push_back(esriFieldTypeDouble);
      break;
    case MULTNAMES_TABLE:
      name = "MultNames";
      vN.Add(CComBSTR("MultID"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("MultName"));        vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("ArrayMult"));       vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Function"));        vT.push_back(esriFieldTypeString);
      break;
    case ZONES_TABLE:
      name = "Zones";
      vN.Add(CComBSTR("ZoneID"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("IJ"));              vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("IZON"));            vT.push_back(esriFieldTypeInteger);
      break;
    case ZONENAMES_TABLE:
      name = "ZoneNames";
      vN.Add(CComBSTR("ZoneID"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("ZoneName"));        vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("ArrayMult"));       vT.push_back(esriFieldTypeInteger);
      break;
    case PARINSTANCES_TABLE:
      name = "ParInstances";
      vN.Add(CComBSTR("ParInstID"));       vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("InstName"));        vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("ParamID"));         vT.push_back(esriFieldTypeInteger);
      break;
    case PARINSTSP_TABLE:
      name = "ParInstSP";
      vN.Add(CComBSTR("ParInstID"));       vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      break;

    // =========================================================================
    // Observations
    case OBSVARS_TABLE:
      name = "OBSVars";
      vN.Add(CComBSTR("OUTNAM"));          vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("ISCALS"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("TOMULTH"));         vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("EVH"));             vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("IUHOBSV"));         vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("HOBDRY"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("TOMULTGB"));        vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("EVFGB"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("IUGBOBSV"));        vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("TOMULTDR"));        vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("EVFDR"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("IUDROBSV"));        vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("TOMULTRV"));        vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("EVFRV"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("IURVOBSV"));        vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("TOMULTCH"));        vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("EVFCH"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("IUCHOBSV"));        vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("TOMULTST"));        vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("EVFST"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("IUSTOBSV"));        vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("TOMULTDT"));        vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("EVFDT"));           vT.push_back(esriFieldTypeDouble);
      break;
    case HOB_TABLE:
      name = "HOB";
      vN.Add(CComBSTR("HOBID"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("OBSNAM"));          vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("IJ"));              vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("ROFF"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("COFF"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("ITT"));             vT.push_back(esriFieldTypeSmallInteger);
      break;
    case HOBLAYERS_TABLE:
      name = "HOBLayers";
      vN.Add(CComBSTR("HOBID"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Layer"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("PR"));              vT.push_back(esriFieldTypeDouble);
      break;
    case HOBTIMES_TABLE:
      name = "HOBTimes";
      vN.Add(CComBSTR("HOBID"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("OBSNAM"));          vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("IREFSP"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("TOFFSET"));         vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("HOBS"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("HSIM"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("STATh"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("STATdd"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("STATFLAG"));        vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("PLOTSYMBOL"));      vT.push_back(esriFieldTypeInteger);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(CMDMDomains::Instance().GetSTATFLAGDomain());
      vD.push_back(NULL);
      break;
    case FLOB_TABLE:
      name = "FLOB";
      vN.Add(CComBSTR("FLOBID"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("FLOBType"));        vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("OBSNAM"));          vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("IREFSP"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("TOFFSET"));         vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("HOBS"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("HSIM"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("STATISTIC"));       vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("STATFLAG"));        vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("PLOTSYMBOL"));      vT.push_back(esriFieldTypeInteger);
      vD.push_back(NULL);
      vD.push_back(CMDMDomains::Instance().GetFLOBTypeDomain());
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      break;
    case FLOBFACTORS_TABLE:
      name = "FLOBFactors";
      vN.Add(CComBSTR("FLOBID"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("IJK"));             vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Factor"));          vT.push_back(esriFieldTypeDouble);
      break;

    // =========================================================================
    // SEN, PES Processes
    case SEN_TABLE:
      name = "SEN";
      vN.Add(CComBSTR("ISENALL"));         vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("IUHEAD"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("IPRINTS"));         vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("ISENSU"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("ISENPU"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("ISENFM"));          vT.push_back(esriFieldTypeSmallInteger);
      break;
    case PES_TABLE:
      name = "PES";
      vN.Add(CComBSTR("MAXITER"));         vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("MAXCHANGE"));       vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("TOL"));             vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("SOSC"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("IBEFLG"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("IYCFLG"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("IOSTAR"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("NOPT"));            vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("NFIT"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SOSR"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("RMAR"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("RMARM"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("IAP"));             vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("IPRCOV"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("IPRINT"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("LPRINT"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("CSA"));             vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("FCONV"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("LASTX"));           vT.push_back(esriFieldTypeDouble);
      break;

    // =========================================================================
    // Output
    case OUTPUTFILES_TABLE:
      name = "OutputFiles";
      vN.Add(CComBSTR("FileType"));        vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("Description"));     vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("Format"));          vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("Compress"));        vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("File"));            vT.push_back(esriFieldTypeBlob);
      vN.Add(CComBSTR("Path"));            vT.push_back(esriFieldTypeString);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(CMDMDomains::Instance().GetOutputFormatDomain());
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(NULL);
      break;
    case OUTPUTTIME_TABLE:
      name = "OutputTime";
      vN.Add(CComBSTR("TimeID"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SPNum"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("TSNum"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("TotalTime"));       vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("PeriodTime"));      vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AbsoluteTime"));    vT.push_back(esriFieldTypeDate);
      break;
    case OUTPUTHEAD_TABLE:
      name = "OutputHead";
      vN.Add(CComBSTR("TimeID"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("IJK"));             vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Head"));            vT.push_back(esriFieldTypeDouble);
      break;
    case OUTPUTDRAWDOWN_TABLE:
      name = "OutputDrawdown";
      vN.Add(CComBSTR("TimeID"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("IJK"));             vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Drawdown"));        vT.push_back(esriFieldTypeDouble);
      break;
    case OUTPUTFLOW_TABLE:
      name = "OutputFlow";
      vN.Add(CComBSTR("TimeID"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("IJK"));             vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("FlowType"));        vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("Flow"));            vT.push_back(esriFieldTypeDouble);
      vD.push_back(NULL);
      vD.push_back(NULL);
      vD.push_back(CMDMDomains::Instance().GetOutputFlowTypeDomain());
      vD.push_back(NULL);
      break;

    // =========================================================================
    // Hydrogeologic Unit Flow (HUF2)
    case HUFVARS_TABLE:
      name = "HUFVars";
      vN.Add(CComBSTR("HDRY"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("NHUF"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("IOHUFHEADS"));      vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("IOHUFFLOWS"));      vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("WETFCT"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("IWETIT"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("IHDWET"));          vT.push_back(esriFieldTypeSmallInteger);
      break;

    case HUFLAYERS_TABLE:
      name = "HUFLayers";
      vN.Add(CComBSTR("Layer"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("LTHUF"));           vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("LAYWT"));           vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("AM_WETDRY"));       vT.push_back(esriFieldTypeDouble);
      break;

    case HUFWETDRY_TABLE:
      name = "HUFWetDry";
      vN.Add(CComBSTR("IJK"));             vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("WETDRY"));          vT.push_back(esriFieldTypeDouble);
      break;

    case HUFUNITS_TABLE:
      name = "HUFUnits";
      vN.Add(CComBSTR("HGUID"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("HGUNAME"));         vT.push_back(esriFieldTypeString);
      vN.Add(CComBSTR("HGUHANI"));         vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("HGUVANI"));         vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_Top"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_Thick"));        vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("PRINTCODE_HK"));    vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("PRINTCODE_HANI"));  vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("PRINTCODE_VK"));    vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("PRINTCODE_Ss"));    vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("PRINTCODE_Sy"));    vT.push_back(esriFieldTypeSmallInteger);
      break;

    case HUFUNITPROPS_TABLE:
      name = "HUFUnitProps";
      vN.Add(CComBSTR("IJ"));              vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("HGUID"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Top"));             vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Thick"));           vT.push_back(esriFieldTypeDouble);
      break;

    // =========================================================================
    // Stream (STR6)
    case STRVARS_TABLE:
      name = "STRVars";
      vN.Add(CComBSTR("ICALC"));           vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("CONST"));           vT.push_back(esriFieldTypeDouble);
      break;

    case STRITRIB_TABLE:
      name = "STRItrib";
      vN.Add(CComBSTR("Seg"));             vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Itrib"));           vT.push_back(esriFieldTypeInteger);
      break;

    case STRIUPSEG_TABLE:
      name = "STRIupseg";
      vN.Add(CComBSTR("Seg"));             vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Iupseg"));          vT.push_back(esriFieldTypeInteger);
      break;

    case STRREACHES_TABLE:
      name = "STRReaches";
      vN.Add(CComBSTR("IJK"));             vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Seg"));             vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Reach"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("Flow"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Stage"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Cond"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Sbot"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Stop"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Width"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Slope"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Rough"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("Condfact"));        vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("SourceID"));        vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("LinearScale"));     vT.push_back(esriFieldTypeDouble);
      break;

    // =========================================================================
    // Streamflow-Routing (SFR2)
    case SFRVARS_TABLE:
      name = "SFRVars";
      vN.Add(CComBSTR("CONST"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("DLEAK"));           vT.push_back(esriFieldTypeDouble);
      break;

    case SFRREACHES_TABLE:
      name = "SFRReaches";
      vN.Add(CComBSTR("IJK"));             vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("ISEG"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("IREACH"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("RCHLEN"));          vT.push_back(esriFieldTypeDouble);
      break;

    case SFRFLOWTAB_TABLE:
      name = "SFRFlowTab";
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("NSEG"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("FLOWTAB"));         vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("DPTHTAB"));         vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("WDTHTAB"));         vT.push_back(esriFieldTypeDouble);
      break;

    case SFRSEGMENTS_TABLE:
      name = "SFRSegments";
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("NSEG"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("ICALC"));           vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("OUTSEG"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("IUPSEG"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("IPRIOR"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("NSTRPTS"));         vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("FLOW"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("RUNOFF"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("ETSW"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("PPTSW"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("ROUGHCH"));         vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("ROUGHBK"));         vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("CDPTH"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("FDPTH"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AWDTH"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("BWDTH"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("HCOND1"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("THICKM1"));         vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("ELEVUP"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("WIDTH1"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("DEPTH1"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("HCOND2"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("THICKM2"));         vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("ELEVDN"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("WIDTH2"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("DEPTH2"));          vT.push_back(esriFieldTypeDouble);
      break;

    case SFRXSECT_TABLE:
      name = "SFRXsect";
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("NSEG"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("XCPT1"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("ZCPT1"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("XCPT2"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("ZCPT2"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("XCPT3"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("ZCPT3"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("XCPT4"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("ZCPT4"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("XCPT5"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("ZCPT5"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("XCPT6"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("ZCPT6"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("XCPT7"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("ZCPT7"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("XCPT8"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("ZCPT8"));           vT.push_back(esriFieldTypeDouble);
      break;

    // =========================================================================
    // Gaging Station File (GAGE)
    case GAGSFR_TABLE:
      name = "GAGSFR";
      vN.Add(CComBSTR("GAGESEG"));         vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("GAGERCH"));         vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("UNIT"));            vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("OUTTYPE"));         vT.push_back(esriFieldTypeSmallInteger);
      break;

    case GAGLAK_TABLE:
      name = "GAGLAK";
      vN.Add(CComBSTR("LAKE"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("UNIT"));            vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("OUTTYPE"));         vT.push_back(esriFieldTypeSmallInteger);
      break;

    // =========================================================================
    // Lake (LAK3)
    case LAKVARS_TABLE:
      name = "LAKVars";
      vN.Add(CComBSTR("THETA"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("NSSITR"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("SSCNCR"));          vT.push_back(esriFieldTypeDouble);
      break;

    case LAKARRAYS_TABLE:
      name = "LAKArrays";
      vN.Add(CComBSTR("IJK"));             vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("LKARR"));           vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("BDLKNC"));          vT.push_back(esriFieldTypeDouble);
      break;

    case LAKSTAGES_TABLE:
      name = "LAKStages";
      vN.Add(CComBSTR("LakeID"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("STAGES"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("SSMN"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("SSMX"));            vT.push_back(esriFieldTypeDouble);
      break;

    case LAKSUBLAKES_TABLE:
      name = "LAKSublakes";
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SystemID"));        vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("ISUB"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SILLVT"));          vT.push_back(esriFieldTypeDouble);
      break;

    case LAKRATES_TABLE:
      name = "LAKRates";
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("LakeID"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("PRCPLK"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("EVAPLK"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("RNF"));             vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("WTHDRW"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("SSMN"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("SSMX"));            vT.push_back(esriFieldTypeDouble);
      break;

    // =========================================================================
    // Unsaturated Zone Flow (UZF)
    case UZFVARS_TABLE:
      name = "UZFVars";
      vN.Add(CComBSTR("NUZTOP"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("IUZFOPT"));         vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("IRUNFLG"));         vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("IETFLG"));          vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("NTRAIL2"));         vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("NSETS2"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("NUZGAG"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SURFDEP"));         vT.push_back(esriFieldTypeDouble);
      break;

    case UZFARRAYS_TABLE:
      name = "UZFArrays";
      vN.Add(CComBSTR("IJ"));              vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("IUZFBND"));         vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("IRUNBND"));         vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("VKS"));             vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("EPS"));             vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("THTS"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("THTI"));            vT.push_back(esriFieldTypeDouble);
      break;

    case UZFARRAYMULT_TABLE:
      name = "UZFArrayMult";
      vN.Add(CComBSTR("AM_IUZFBND"));      vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("AM_IRUNBND"));      vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("AM_VKS"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_EPS"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_THTS"));         vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_THTI"));         vT.push_back(esriFieldTypeDouble);
      break;

    case UZFGAGES_TABLE:
      name = "UZFGages";
      vN.Add(CComBSTR("IUZROW"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("IUZCOL"));          vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("IFTUNIT"));         vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("IUZOPT"));          vT.push_back(esriFieldTypeSmallInteger);
      break;

    case UZFSTRESSARRAYS_TABLE:
      name = "UZFStressArrays";
      vN.Add(CComBSTR("IJ"));              vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("FINF"));            vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("PET"));             vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("EXTDP"));           vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("EXTWC"));           vT.push_back(esriFieldTypeDouble);
      break;

    case UZFSTRESSARRAYMULT_TABLE:
      name = "UZFStressArrayMult";
      vN.Add(CComBSTR("SPID"));            vT.push_back(esriFieldTypeInteger);
      vN.Add(CComBSTR("AM_FINF"));         vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_PET"));          vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_EXTDP"));        vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("AM_EXTWC"));        vT.push_back(esriFieldTypeDouble);
      vN.Add(CComBSTR("NUZF1"));           vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("NUZF2"));           vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("NUZF3"));           vT.push_back(esriFieldTypeSmallInteger);
      vN.Add(CComBSTR("NUZF4"));           vT.push_back(esriFieldTypeSmallInteger);
      break;

    // =========================================================================
    // Error:  Not a valid table ID
    default:
      rval = 0;
      break;
  }

  // add aux fields to tables
  if (TableID == RIV_TABLE ||
      TableID == CHD_TABLE ||
      TableID == WEL_TABLE ||
      TableID == DRN_TABLE ||
      TableID == GHB_TABLE)
  {
    std::map<CStr, std::set<CStr> >& m(iGetAuxCols());
    CStr tab(name);
    if (m.find(tab) != m.end())
    {
      std::set<CStr>::iterator it, end;
      end = m.find(tab)->second.end();
      it = m.find(tab)->second.begin();
      for (; it != end; it++)
      {
        vN.Add(CComBSTR(*it));  vT.push_back(esriFieldTypeDouble);
      }
    }
  }

  return rval;
}

// *****************************************************************************
// NOTES:  Checks if the name passed in exists in the field definition.  Does
//         not actually read a geodb table to see if that field is there.
// *****************************************************************************
bool CTableDef::FieldExistsInDef (CStr tablename, CStr fieldname)
{
  int i(0);
  int tableid = GetTableIdFromString(tablename);
  if (tableid < 0)
    return false;

  return (FieldExistsInDef(tableid, fieldname));
}

// *****************************************************************************
// NOTES:  Checks if the name passed in exists in the field definition.  Does
//         not actually read a geodb table to see if that field is there.
// *****************************************************************************
bool CTableDef::FieldExistsInDef (int tableid, CStr fieldname)
{
  unsigned long i;

  // Get the field names
  CComSafeArray<BSTR> vNames;
  VEC_FTYPE vTypes;
  VEC_DOMAIN vDomain;
  CComBSTR tName;


  if (GetData(tableid, vNames, vTypes, vDomain, tName)) {
    for (i = 0; i < vNames.GetCount(); ++i) {
      CStr tempname(vNames.GetAt(i));
      if (fieldname.CompareNoCase(tempname) == 0)
        return true;
    }
  }

  return false;
}

// *****************************************************************************
// NOTES:  Returns the ID of the table, based on the table name passed in
//         Returns -1 if it is not valid
// *****************************************************************************
int CTableDef::GetTableIdFromString (CStr name)
{
  ma_tables table;

  if (name.CompareNoCase("MDFGlobals") == 0)
    table = MDFGLOBALS_TABLE;
  else if (name.CompareNoCase("NameFile") == 0)
    table = NAMEFILE_TABLE;
  else if (name.CompareNoCase("DISVars") == 0)
    table = DISVARS_TABLE;
  else if (name.CompareNoCase("DELRC") == 0)
    table = DELRC_TABLE;
  else if (name.CompareNoCase("TopElev") == 0)
    table = TOPELEV_TABLE;
  else if (name.CompareNoCase("BotmElev") == 0)
    table = BOTMELEV_TABLE;
  else if (name.CompareNoCase("DISLayers") == 0)
    table = DISLAYERS_TABLE;
  else if (name.CompareNoCase("StressPeriods") == 0)
    table = STRESSPERIODS_TABLE;
  else if (name.CompareNoCase("BASVars") == 0)
    table = BASICVARS_TABLE;
  else if (name.CompareNoCase("Basic") == 0)
    table = BASIC_TABLE;
  else if (name.CompareNoCase("BasicArrayMult") == 0)
    table = BASIC_MULT_TABLE;
  else if (name.CompareNoCase("OCVars") == 0)
    table = OCVARS_TABLE;
  else if (name.CompareNoCase("OCTS") == 0)
    table = OCTS_TABLE;
  else if (name.CompareNoCase("CBFlags") == 0)
    table = CBFLAGS_TABLE;
  else if (name.CompareNoCase("BCFVars") == 0)
    table = BCFVARS_TABLE;
  else if (name.CompareNoCase("BCFLayers") == 0)
    table = BCFLAYERS_TABLE;
  else if (name.CompareNoCase("BCFProperties") == 0)
    table = BCFPROPERTIES_TABLE;
  else if (name.CompareNoCase("LPFVars") == 0)
    table = LPFVARS_TABLE;
  else if (name.CompareNoCase("LPFLayers") == 0)
    table = LPFLAYERS_TABLE;
  else if (name.CompareNoCase("LPFProperties") == 0)
    table = LPFPROPERTIES_TABLE;
  else if (name.CompareNoCase("RIV") == 0)
    table = RIV_TABLE;
  else if (name.CompareNoCase("DRN") == 0)
    table = DRN_TABLE;
  else if (name.CompareNoCase("CHD") == 0)
    table = CHD_TABLE;
  else if (name.CompareNoCase("GHB") == 0)
    table = GHB_TABLE;
  else if (name.CompareNoCase("WEL") == 0)
    table = WEL_TABLE;
  else if (name.CompareNoCase("HFB6") == 0)
    table = HFB6_TABLE;
  else if (name.CompareNoCase("DRT") == 0)
    table = DRT_TABLE;
  else if (name.CompareNoCase("MNW1Vars") == 0)
    table = MNW1VARS_TABLE;
  else if (name.CompareNoCase("MNW1") == 0)
    table = MNW1_TABLE;
  else if (name.CompareNoCase("RCHVars") == 0)
    table = RCHVARS_TABLE;
  else if (name.CompareNoCase("RCHArrayMult") == 0)
    table = RCHARRAYMULT_TABLE;
  else if (name.CompareNoCase("RCHArrays") == 0)
    table = RCHARRAYS_TABLE;
  else if (name.CompareNoCase("EVTVars") == 0)
    table = EVTVARS_TABLE;
  else if (name.CompareNoCase("EVTArrayMult") == 0)
    table = EVTARRAYMULT_TABLE;
  else if (name.CompareNoCase("EVTArrays") == 0)
    table = EVTARRAYS_TABLE;
  else if (name.CompareNoCase("ETSVars") == 0)
    table = ETSVARS_TABLE;
  else if (name.CompareNoCase("ETSArrays") == 0)
    table = ETSARRAYS_TABLE;
  else if (name.CompareNoCase("ETSArrayMult") == 0)
    table = ETSARRAYMULT_TABLE;
  else if (name.CompareNoCase("ETSSegArrays") == 0)
    table = ETSSEGARRAYS_TABLE;
  else if (name.CompareNoCase("ETSSegArrayMult") == 0)
    table = ETSSEGARRAYMULT_TABLE;
  else if (name.CompareNoCase("SEN") == 0)
    table = SEN_TABLE;
  else if (name.CompareNoCase("PES") == 0)
    table = PES_TABLE;
  else if (name.CompareNoCase("SIP") == 0)
    table = SIP_TABLE;
  else if (name.CompareNoCase("PCG") == 0)
    table = PCG_TABLE;
  else if (name.CompareNoCase("GMG") == 0)
    table = GMG_TABLE;
  else if (name.CompareNoCase("DE4") == 0)
    table = DE4_TABLE;
  else if (name.CompareNoCase("LMG") == 0)
    table = LMG_TABLE;
  else if (name.CompareNoCase("SOR") == 0)
    table = SOR_TABLE;
  else if (name.CompareNoCase("Params") == 0)
    table = PARAMS_TABLE;
  else if (name.CompareNoCase("Clusters") == 0)
    table = CLUSTERS_TABLE;
  else if (name.CompareNoCase("IZ") == 0)
    table = IZ_TABLE;
  else if (name.CompareNoCase("PilotPts") == 0)
    table = PILOTPTS_TABLE;
  else if (name.CompareNoCase("ParInstances") == 0)
    table = PARINSTANCES_TABLE;
  else if (name.CompareNoCase("ParInstSP") == 0)
    table = PARINSTSP_TABLE;
  else if (name.CompareNoCase("Multipliers") == 0)
    table = MULTIPLIERS_TABLE;
  else if (name.CompareNoCase("MultNames") == 0)
    table = MULTNAMES_TABLE;
  else if (name.CompareNoCase("Zones") == 0)
    table = ZONES_TABLE;
  else if (name.CompareNoCase("ZoneNames") == 0)
    table = ZONENAMES_TABLE;
  else if (name.CompareNoCase("OBSVars") == 0)
    table = OBSVARS_TABLE;
  else if (name.CompareNoCase("HOB") == 0)
    table = HOB_TABLE;
  else if (name.CompareNoCase("HOBLayers") == 0)
    table = HOBLAYERS_TABLE;
  else if (name.CompareNoCase("HOBTimes") == 0)
    table = HOBTIMES_TABLE;
  else if (name.CompareNoCase("FLOB") == 0)
    table = FLOB_TABLE;
  else if (name.CompareNoCase("FLOBFactors") == 0)
    table = FLOBFACTORS_TABLE;
  else if (name.CompareNoCase("OutputFiles") == 0)
    table = OUTPUTFILES_TABLE;
  else if (name.CompareNoCase("OutputTime") == 0)
    table = OUTPUTTIME_TABLE;
  else if (name.CompareNoCase("OutputHead") == 0)
    table = OUTPUTHEAD_TABLE;
  else if (name.CompareNoCase("OutputDrawdown") == 0)
    table = OUTPUTDRAWDOWN_TABLE;
  else if (name.CompareNoCase("OutputFlow") == 0)
    table = OUTPUTFLOW_TABLE;
  else if (name.CompareNoCase("HUFVars") == 0)
    table = HUFVARS_TABLE;
  else if (name.CompareNoCase("HUFLayers") == 0)
    table = HUFLAYERS_TABLE;
  else if (name.CompareNoCase("HUFWetDry") == 0)
    table = HUFWETDRY_TABLE;
  else if (name.CompareNoCase("HUFUnits") == 0)
    table = HUFUNITS_TABLE;
  else if (name.CompareNoCase("HUFUnitProps") == 0)
    table = HUFUNITPROPS_TABLE;
  else if (name.CompareNoCase("STRVars") == 0)
    table = STRVARS_TABLE;
  else if (name.CompareNoCase("STRItrib") == 0)
    table = STRITRIB_TABLE;
  else if (name.CompareNoCase("STRIupseg") == 0)
    table = STRIUPSEG_TABLE;
  else if (name.CompareNoCase("STRReaches") == 0)
    table = STRREACHES_TABLE;
  else if (name.CompareNoCase("SFRVars") == 0)
    table = SFRVARS_TABLE;
  else if (name.CompareNoCase("SFRReaches") == 0)
    table = SFRREACHES_TABLE;
  else if (name.CompareNoCase("SFRFlowTab") == 0)
    table = SFRFLOWTAB_TABLE;
  else if (name.CompareNoCase("SFRSegments") == 0)
    table = SFRSEGMENTS_TABLE;
  else if (name.CompareNoCase("SFRXsect") == 0)
    table = SFRXSECT_TABLE;
  else if (name.CompareNoCase("GAGSFR") == 0)
    table = GAGSFR_TABLE;
  else if (name.CompareNoCase("GAGLAK") == 0)
    table = GAGLAK_TABLE;
  else if (name.CompareNoCase("LAKVars") == 0)
    table = LAKVARS_TABLE;
  else if (name.CompareNoCase("LAKArrays") == 0)
    table = LAKARRAYS_TABLE;
  else if (name.CompareNoCase("LAKStages") == 0)
    table = LAKSTAGES_TABLE;
  else if (name.CompareNoCase("LAKSublakes") == 0)
    table = LAKSUBLAKES_TABLE;
  else if (name.CompareNoCase("LAKRates") == 0)
    table = LAKRATES_TABLE;
  else if (name.CompareNoCase("UZFVars") == 0)
    table = UZFVARS_TABLE;
  else if (name.CompareNoCase("UZFArrays") == 0)
    table = UZFARRAYS_TABLE;
  else if (name.CompareNoCase("UZFArrayMult") == 0)
    table = UZFARRAYMULT_TABLE;
  else if (name.CompareNoCase("UZFGages") == 0)
    table = UZFGAGES_TABLE;
  else if (name.CompareNoCase("UZFStressArrays") == 0)
    table = UZFSTRESSARRAYS_TABLE;
  else if (name.CompareNoCase("UZFStressArrayMult") == 0)
    table = UZFSTRESSARRAYMULT_TABLE;
  else
    return -1;

  return (int)table;
}

//------------------------------------------------------------------------------
#ifdef CXX_TEST

#include <TableDef.t.h>

////////////////////////////////////////////////////////////////////////////////
//
// class CTableDefT
//
////////////////////////////////////////////////////////////////////////////////
//------------------------------------------------------------------------------
void CTableDefT::testGetData ()
{
  //TS_FAIL("CTableDefT::testGetData");

  CTableDef tableDef;
  CComSafeArray<BSTR> vN;
  VEC_FTYPE vT;
  VEC_DOMAIN vD;
  CComBSTR name;

  // We test a couple of these to make sure things work, but we don't test all

  TS_ASSERT_EQUALS(tableDef.GetData(MDFGLOBALS_TABLE, vN, vT, vD, name), 1);
  TS_ASSERT_EQUALS(vN.GetCount(), 3);
  TS_ASSERT_EQUALS(vN[0], "MODFLOW_Version");
  TS_ASSERT_EQUALS(vN[1], "DataModelVersion");
  TS_ASSERT_EQUALS(vN[2], "NameFilePath");
  TS_ASSERT_EQUALS((int)vT.size(), 3);
  TS_ASSERT_EQUALS(vT[0], esriFieldTypeString);
  TS_ASSERT_EQUALS(vT[1], esriFieldTypeString);
  TS_ASSERT_EQUALS(vT[2], esriFieldTypeString);
  TS_ASSERT_EQUALS((int)vD.size(), 0);

  TS_ASSERT_EQUALS(tableDef.GetData(OUTPUTFLOW_TABLE, vN, vT, vD, name), 1);
  TS_ASSERT_EQUALS(vN.GetCount(), 4);
  TS_ASSERT_EQUALS(vN[0], "TimeID");
  TS_ASSERT_EQUALS(vN[1], "IJK");
  TS_ASSERT_EQUALS(vN[2], "FlowType");
  TS_ASSERT_EQUALS(vN[3], "Flow");
  TS_ASSERT_EQUALS((int)vT.size(), 4);
  TS_ASSERT_EQUALS(vT[0], esriFieldTypeInteger);
  TS_ASSERT_EQUALS(vT[1], esriFieldTypeInteger);
  TS_ASSERT_EQUALS(vT[2], esriFieldTypeSmallInteger);
  TS_ASSERT_EQUALS(vT[3], esriFieldTypeDouble);
  TS_ASSERT_EQUALS((int)vD.size(), 4);
  TS_ASSERT_EQUALS(vD[0], NULL);
  TS_ASSERT_EQUALS(vD[1], NULL);
  TS_ASSERT_EQUALS(vD[2], NULL);
  TS_ASSERT_EQUALS(vD[3], NULL);
}
//------------------------------------------------------------------------------
void CTableDefT::testFieldExistsInDef ()
{
  //TS_FAIL("CTableDefT::testFieldExistsInDef");

  CTableDef tableDef;

  // We test a couple of these to make sure things work, but we don't test all

  TS_ASSERT(tableDef.FieldExistsInDef("MDFGlobals", "MODFLOW_Version"));
  TS_ASSERT(tableDef.FieldExistsInDef("MDFGlobals", "DataModelVersion"));

  TS_ASSERT(tableDef.FieldExistsInDef("OutputFlow", "TimeID"));
  TS_ASSERT(tableDef.FieldExistsInDef("OutputFlow", "IJK"));
  TS_ASSERT(tableDef.FieldExistsInDef("OutputFlow", "FlowType"));
  TS_ASSERT(tableDef.FieldExistsInDef("OutputFlow", "Flow"));
}
//------------------------------------------------------------------------------
void CTableDefT::testGetTableIdFromString ()
{
  //TS_FAIL("CTableDefT::testGetTableIdFromString");

  CTableDef tableDef;

  // We test a couple of these to make sure things work, but we don't test all

  TS_ASSERT_EQUALS(tableDef.GetTableIdFromString(""), -1);
  TS_ASSERT_EQUALS(tableDef.GetTableIdFromString("MDFGlobals"), MDFGLOBALS_TABLE);
  TS_ASSERT_EQUALS(tableDef.GetTableIdFromString("OutputFlow"), OUTPUTFLOW_TABLE);
}

#endif

