// *****************************************************************************
// NOTES:  This code is for creating new geodatabases, tables, etc. if they
//         do not already exist.
// *****************************************************************************

#include <sys/stat.h> 

#include "CreateGDB.h"
#include "TableDef.h"
#include "MDMDomains.h"

// *****************************************************************************
CCreateGDB::CCreateGDB(void)
{
  // Register for COM to use the ESRI stuff
  ::CoInitialize(0);

  m_szPath = "";
  m_szFile = "";
  m_Personal = 1;
}

// *****************************************************************************
CCreateGDB::~CCreateGDB(void)
{
  // Unregister COM
  ::CoUninitialize();
}

// *****************************************************************************
// NOTES:  Creates a new workspace factory, necessary for starting a new
//         geodatabase.
// *****************************************************************************
HRESULT CCreateGDB::CreateWorkspaceFactory (IWorkspaceFactory **ppFact)
{
  HRESULT hr;
  IWorkspaceFactoryPtr ipWorkspaceFactory;

  // Make either an Access Personal Geodatabase or File Geodatabase
  if (m_Personal == 1)
    hr = ipWorkspaceFactory.CreateInstance(CLSID_AccessWorkspaceFactory);
  else
    hr = ipWorkspaceFactory.CreateInstance(CLSID_FileGDBWorkspaceFactory);
  if (FAILED(hr)) return hr;

  (*ppFact) = ipWorkspaceFactory;
  if (*ppFact)
    (*ppFact)->AddRef();

  return(hr);
}

// *****************************************************************************
// NOTES:  Creates a new workspace (i.e. Geodatabase).  The path to this
//         workspace should have been previously set.
// *****************************************************************************
HRESULT CCreateGDB::CreateWorkspace (IWorkspaceFactoryPtr ipFact,
                                     IWorkspace **ppWorkspace)
{
  IWorkspaceNamePtr  ipWorkspaceName;
  INamePtr           ipName;
  IUnknownPtr        ipUnk;
  HRESULT            hr;
  IPropertySetPtr    ipProperty;

  // Create the workspace property set
  hr = ipProperty.CreateInstance(CLSID_PropertySet);

  std::string strFull;

  strFull = m_szStdPath + "\\" + m_szStdFile;

  if (!FileExists(strFull))
  {
    // Using the workspace factory, create a new workspace, with the properties
    // of it stored in the workspace name interface
    hr = ipFact->Create(m_szPath, m_szFile, ipProperty, 0,
                        &ipWorkspaceName);

    // Check to see if we were able to create the workspace or not
    if (FAILED(hr))
      return(hr);

    // Open the workspace
    ipName = ipWorkspaceName;
    hr = ipName->Open(&ipUnk);
    IWorkspacePtr ipTempWorkspace;
    ipTempWorkspace = ipUnk;
    (*ppWorkspace) = ipTempWorkspace;
  }
  else
  {
    // Try opening the workspace from the file name
    IWorkspacePtr ipTempWorkspace;
    CComBSTR bstrFile;
    bstrFile = m_szPath;
    bstrFile += "\\";
    bstrFile += m_szFile;
    hr = ipFact->OpenFromFile(bstrFile, NULL, ppWorkspace);
    if (FAILED(hr))
      return(hr);
  }

  if (*ppWorkspace) {
    (*ppWorkspace)->AddRef();
    return(S_OK);
  }
  return(E_FAIL);
}

// *****************************************************************************
// NOTES:  Starts an edit session.
// *****************************************************************************
HRESULT CCreateGDB::StartEditing (IWorkspacePtr ipWorkspace)
{
  HRESULT hr = E_FAIL;
  IWorkspaceEditPtr ipWorkspaceEdit(ipWorkspace);

  // Start editing, disabling the undo/redo logging
  if (ipWorkspaceEdit != 0)
    hr = ipWorkspaceEdit->StartEditing(VARIANT_FALSE);

  return(hr);
}

// *****************************************************************************
// NOTES:  Ends an edit session.
// *****************************************************************************
HRESULT CCreateGDB::StopEditing (IWorkspacePtr ipWorkspace)
{
  HRESULT hr;
  VARIANT_BOOL bEditing;
  IWorkspaceEditPtr ipWorkspaceEdit(ipWorkspace);

  // Make sure that we were really editing at first
  hr = ipWorkspaceEdit->IsBeingEdited(&bEditing);

  // If editing, stop editing, saving our edits
  if (bEditing == VARIANT_TRUE)
    hr = ipWorkspaceEdit->StopEditing(VARIANT_TRUE);

  return(hr);
}

// *****************************************************************************
// NOTES:  Sets up the field information for a feature class, including the
//         geometry info and attributes
// *****************************************************************************
HRESULT CCreateGDB::CreateFieldsEditFeature (esriGeometryType geom,
                                             IFieldsEdit **ppFE,
                                             ISpatialReferencePtr ipSRef/*= 0*/,
                                             bool bHasZ /* = false */,
                                             bool bHasM /* = true */)
{
  HRESULT           hr;
  IFieldsEditPtr    ipFieldsEdit;
  IFieldEditPtr     ipFieldEdit;
  long              numpoints;
  VARIANT_BOOL      bZ, bM;

  bZ = (bHasZ) ? VARIANT_TRUE : VARIANT_FALSE;
  bM = (bHasM) ? VARIANT_TRUE : VARIANT_FALSE;

  hr = ipFieldsEdit.CreateInstance(CLSID_Fields);

  // Create the Object ID Field (required)
  hr = ipFieldEdit.CreateInstance(CLSID_Field);
  hr = ipFieldEdit->put_Name(L"OID");
  hr = ipFieldEdit->put_Type(esriFieldTypeOID);
  hr = ipFieldEdit->put_AliasName(L"Object ID");
  hr = ipFieldEdit->put_IsNullable(VARIANT_FALSE);
  hr = ipFieldsEdit->AddField(ipFieldEdit);

  // Create the geometry definition
  IGeometryDefEditPtr  ipGeomDef;
  hr = ipGeomDef.CreateInstance(CLSID_GeometryDef);
  switch (geom) {
    case esriGeometryPoint:
      numpoints = 1;
      break;
    case esriGeometryPolyline:
      numpoints = 5;
      break;
    case esriGeometryPolygon:
      numpoints = 20;
      break;
    default:
      numpoints = 1;
      break;
  }
  hr = ipGeomDef->put_AvgNumPoints(numpoints);
  hr = ipGeomDef->put_GeometryType(geom);
  hr = ipGeomDef->put_GridCount(1);
  hr = ipGeomDef->put_GridSize(0, 100);
  hr = ipGeomDef->put_HasM(bM);
  hr = ipGeomDef->put_HasZ(bZ);
  if (ipSRef != 0)
    hr = ipGeomDef->putref_SpatialReference(ipSRef);

  // Create the Shape (geometry) field (required)
  hr = ipFieldEdit.CreateInstance(CLSID_Field);
  hr = ipFieldEdit->put_Name(L"Shape");
  hr = ipFieldEdit->put_Type(esriFieldTypeGeometry);
  hr = ipFieldEdit->put_IsNullable(VARIANT_TRUE);
  hr = ipFieldEdit->put_Editable(VARIANT_TRUE);
  hr = ipFieldEdit->put_AliasName(L"Shape");
  hr = ipFieldEdit->putref_GeometryDef(ipGeomDef);
  hr = ipFieldsEdit->AddField(ipFieldEdit);

  *ppFE = ipFieldsEdit;
  if (*ppFE)
    (*ppFE)->AddRef();

  return(hr);
}

// *****************************************************************************
// NOTES:  Sets up the field information for a table
// *****************************************************************************
HRESULT CCreateGDB::CreateFieldsEditTable (IFieldsEdit **ppFE)
{
  HRESULT           hr;
  IFieldsEditPtr    ipFieldsEdit;
  IFieldEditPtr     ipFieldEdit;

  hr = ipFieldsEdit.CreateInstance(CLSID_Fields);

  // Create the Object ID Field (required)
  hr = ipFieldEdit.CreateInstance(CLSID_Field);
  hr = ipFieldEdit->put_Name(L"OID");
  hr = ipFieldEdit->put_Type(esriFieldTypeOID);
  hr = ipFieldEdit->put_AliasName(L"Object ID");
  hr = ipFieldEdit->put_IsNullable(VARIANT_FALSE);
  hr = ipFieldsEdit->AddField(ipFieldEdit);

  *ppFE = ipFieldsEdit;
  if (*ppFE)
    (*ppFE)->AddRef();

  return(hr);
}

// *****************************************************************************
// NOTES:  Adds a single field to an IFieldsEdit, with the default parameters
//         based on the type of field to be added.  This is not for a field
//         type for features (Point, Polyline, etc.).
// *****************************************************************************
HRESULT CCreateGDB::AddFieldToFieldsEdit (IFieldsEditPtr ipFE, CComBSTR name,
                                          esriFieldType type,
                                          VARIANT_BOOL isNullable)
{
  HRESULT        hr = S_OK;
  IFieldEditPtr  ipFieldEdit;

  // Create a new field
  hr = ipFieldEdit.CreateInstance(CLSID_Field);
  if (FAILED(hr))
    return hr;

  // Add its name, type, etc.
  hr = ipFieldEdit->put_Name(name);
  if (FAILED(hr))
    return hr;
  hr = ipFieldEdit->put_Type(type);
  if (FAILED(hr))
    return hr;
  hr = ipFieldEdit->put_AliasName(name);
  hr = ipFieldEdit->put_IsNullable(isNullable);
  hr = ipFE->AddField(ipFieldEdit);

  return hr;
}

// *****************************************************************************
// NOTES:  Creates a new table of the given name in the specified workspace.
// *****************************************************************************
HRESULT CCreateGDB::CreateNewTable (CComBSTR szName,
                                    IFeatureWorkspacePtr ipFW,
                                    IFieldsEditPtr ipFields,
                                    IUIDPtr ipUID, IUIDPtr ipEXTUID,
                                    ITable **ppTable)
{
  HRESULT hr;

  // Create the table in the feature class
  hr = ipFW->CreateTable(szName, ipFields, ipUID, ipEXTUID, L"", ppTable);
  if (FAILED(hr))
    return hr;

  if (*ppTable)
    (*ppTable)->AddRef();
  return(hr);
}

// *****************************************************************************
// NOTES:  Returns a new feature class.
//         Some feature class names, such as "General" are invalid.  This
//         function will add a "_" on the end if it gets the error about a bad
//         name.  See comments for additional info
// *****************************************************************************
HRESULT CCreateGDB::CreateNewFeatureClass (CComBSTR szName,
                                           IFeatureDatasetPtr ipFD,
                                           IFieldsEditPtr ipFields,
                                           IUIDPtr ipUID,
                                           IFeatureClass **ppFC)
{
  HRESULT   hr;
  CComBSTR  szBlank, szShape;

  // Set the defaults for some of the parameters
  szBlank = "";
  szShape = "Shape";

  if (ipFD == 0)
    return E_FAIL;

  // Some string values, such as "General," give ArcObjects problems when
  // you try to create a feature class.  You will get a FDO_E_TABLE_NOT_FOUND
  // or -2147220655 error.  "General" is a known problem, there may be others.
  // So, if the name it "General", adding a "_" will make it work.
  // "Parameters" and "Top" are also invalid names.
  hr = ipFD->CreateFeatureClass(szName, ipFields, ipUID, 0, esriFTSimple,
                                szShape, szBlank, ppFC);

  if (hr == FDO_E_TABLE_NOT_FOUND) {
    // If we accidentally run into another problem string other than "General",
    // try the same fix by adding another "_" to the name.
    szName += "_";
    hr = ipFD->CreateFeatureClass(szName, ipFields, ipUID, 0, esriFTSimple,
                                  szShape, szBlank, ppFC);
    if (FAILED(hr))
      return hr;
  }

  if (*ppFC)
    (*ppFC)->AddRef();
  return hr;
}

// *****************************************************************************
// NOTES:  Returns a new feature class.
//         Some feature class names, such as "General" are invalid.  This
//         function will add a "_" on the end if it gets the error about a bad
//         name.  See comments for additional info.
//         This Feature Class is not stored in a feature dataset, but rather
//         in the root of the GDB.
// *****************************************************************************
HRESULT CCreateGDB::CreateNewFeatureClassWksp (CComBSTR szName,
                                               IFeatureWorkspacePtr ipFW,
                                               IFieldsEditPtr ipFields,
                                               IUIDPtr ipUID,
                                               IFeatureClass **ppFC)
{
  HRESULT   hr;
  CComBSTR  szBlank, szShape;

  // Set the defaults for some of the parameters
  szBlank = "";
  szShape = "Shape";

  if (ipFW == 0)
    return E_FAIL;

  // Some string values, such as "General," give ArcObjects problems when
  // you try to create a feature class.  You will get a FDO_E_TABLE_NOT_FOUND
  // or -2147220655 error.  "General" is a known problem, there may be others.
  // So, if the name it "General", adding a "_" will make it work.
  // "Parameters" and "Top" are also invalid names.
  hr = ipFW->CreateFeatureClass(szName, ipFields, ipUID, 0, esriFTSimple,
                                szShape, szBlank, ppFC);

  if (hr == FDO_E_TABLE_NOT_FOUND) {
    // If we accidentally run into another problem string other than "General",
    // try the same fix by adding another "_" to the name.
    szName += "_";
    hr = ipFW->CreateFeatureClass(szName, ipFields, ipUID, 0, esriFTSimple,
                                  szShape, szBlank, ppFC);
    if (FAILED(hr))
      return hr;
  }

  if (*ppFC)
    (*ppFC)->AddRef();
  return hr;
}

// *****************************************************************************
// NOTES:  Creates a feature class with just the bare necessities, excluding
//         geometry, the feature dataset passed in
// *****************************************************************************
HRESULT CCreateGDB::CreateBasicFeatureClass (CComBSTR szName,
                                             esriGeometryType geom,
                                             IFeatureDatasetPtr
                                             ipFeatureDataset,
                                             IUIDPtr ipUID,
                                             IFeatureClass **ipFC)
{
  HRESULT           hr;
  IFieldsEditPtr    ipFieldsEdit;

  // Create a new feature class with basic fields
  hr = CreateFieldsEditFeature(geom, &ipFieldsEdit);
  hr = CreateNewFeatureClass(szName, ipFeatureDataset, ipFieldsEdit,
                             ipUID, ipFC);

  return hr;
}

// *****************************************************************************
// NOTES:  Creates a feature class with just the bare necessities, excluding
//         geometry, the feature dataset passed in.
//         This creates the feature class in the GDB root, as opposed to
//         a feature dataset.
// *****************************************************************************
HRESULT CCreateGDB::CreateBasicFeatureClassWksp (CComBSTR szName,
                                                 esriGeometryType geom,
                                                 IFeatureWorkspacePtr
                                                 ipWorkspace,
                                                 IUIDPtr ipUID,
                                                 IFeatureClass **ipFC)
{
  HRESULT               hr;
  IFieldsEditPtr        ipFieldsEdit;
  ISpatialReferencePtr  ipSpatialRef;

  // Create a dummy spatial reference
  hr = ipSpatialRef.CreateInstance(CLSID_UnknownCoordinateSystem);
  if (FAILED(hr)) return hr;
  hr = ipSpatialRef->SetDomain(-4.5e11, 4.5e11, -4.5e11, 4.5e11);
  if (FAILED(hr)) return hr;
  hr = ipSpatialRef->SetFalseOriginAndUnits(-4.5e11, -4.5e11, 10000.0);
  if (FAILED(hr)) return hr;
  hr = ipSpatialRef->SetMDomain(0.0, 1.0);
  if (FAILED(hr)) return hr;
  hr = ipSpatialRef->SetMFalseOriginAndUnits(0.0, 100.0);
  if (FAILED(hr)) return hr;
  hr = ipSpatialRef->SetZDomain(-1.0e5, 9.0e11);
  if (FAILED(hr)) return hr;
  hr = ipSpatialRef->SetZFalseOriginAndUnits(-1.0e5, 10000.0);

  // Create a new feature class with basic fields
  hr = CreateFieldsEditFeature(geom, &ipFieldsEdit, ipSpatialRef, true, true);
  hr = CreateNewFeatureClassWksp(szName, ipWorkspace, ipFieldsEdit,
                                 ipUID, ipFC);

  return hr;
}

// *****************************************************************************
// NOTES:  Creates a new field based on everything passed in
//         lDefault can be -1, to indicate no default value will be added
// *****************************************************************************
HRESULT CCreateGDB::CreateField (CComBSTR szName, esriFieldType eType,
                                 bool bNullable, long lPrecision, long lLength,
                                 long lDefault, long lScale,
                                 IDomainPtr ipDomain, IFieldEdit **ppFE)
{
  HRESULT        hr;
  IFieldEditPtr  ipFieldEdit;
  CComVariant    vDefault(lDefault);

  // Create the field interface, set name, type, etc.
  hr = ipFieldEdit.CreateInstance(CLSID_Field);
  if (FAILED(hr))
    return hr;
  hr = ipFieldEdit->put_Name(szName);
  hr = ipFieldEdit->put_Type(eType);
  hr = ipFieldEdit->put_AliasName(szName);
  if (bNullable)
    hr = ipFieldEdit->put_IsNullable(VARIANT_TRUE);
  else
    hr = ipFieldEdit->put_IsNullable(VARIANT_FALSE);

  // Add a few additional attributes based on the type of field
  switch (eType) {
    case esriFieldTypeInteger:
      hr = ipFieldEdit->put_Precision(lPrecision);
      if (lDefault != -1)
        hr = ipFieldEdit->put_DefaultValue(vDefault);
      break;

    case esriFieldTypeSmallInteger:
      hr = ipFieldEdit->put_Precision(lPrecision);
      if (lDefault != -1)
        hr = ipFieldEdit->put_DefaultValue(vDefault);
      break;

    case esriFieldTypeDouble:
      hr = ipFieldEdit->put_Precision(lPrecision);
      if (lDefault != -1)
        hr = ipFieldEdit->put_DefaultValue(vDefault);
      hr = ipFieldEdit->put_Scale(lScale);
      break;

    case esriFieldTypeString:
      hr = ipFieldEdit->put_Length(lLength);
      break;
  }

  // Add a domain if one was passed in
  if (ipDomain != 0) {
    hr = ipFieldEdit->putref_Domain(ipDomain);
    if (FAILED(hr))
      return hr;
  }

  (*ppFE) = ipFieldEdit;
  if (*ppFE) {
    (*ppFE)->AddRef();
    return hr;
  }

  return E_FAIL;
}

// *****************************************************************************
// NOTES:  Creates globally unique identifiers objects for tables
// *****************************************************************************
HRESULT CCreateGDB::CreateTableUID (IUID **ppUID, IUID **ppEXTUID)
{
  HRESULT  hr;
  IUIDPtr  ipUid, ipUid2;

  // Don't modify this:
  CComVariant vVariant("{7A566981-C114-11D2-8A28-006097AFF44E}");

  hr = ipUid.CreateInstance(CLSID_UID);
  if (FAILED(hr)) return hr;
  hr = ipUid->put_Value(vVariant);

  (*ppUID) = ipUid;
  if (*ppUID)
    (*ppUID)->AddRef();

  hr = ipUid2.CreateInstance(CLSID_UID);
  if (FAILED(hr)) return hr;
  hr = ipUid2->put_Value(vVariant);

  (*ppEXTUID) = ipUid;
  if (*ppEXTUID)
    (*ppEXTUID)->AddRef();

  return hr;
}

// *****************************************************************************
// NOTES:  Creates globally unique identifiers objects for features
// *****************************************************************************
HRESULT CCreateGDB::CreateFeatureUID (IUID **ppUID)
{
  HRESULT   hr;
  IUIDPtr   ipUid;

  // Don't modify this:
  CComVariant vVariant("{52353152-891A-11D0-BEC6-00805F7C4268}");

  hr = ipUid.CreateInstance(CLSID_UID);
  if (FAILED(hr)) return(hr);
  hr = ipUid->put_Value(vVariant);

  (*ppUID) = ipUid;
  if (*ppUID)
    (*ppUID)->AddRef();

  return(hr);
}

// *****************************************************************************
// NOTES:  Creates a new Geodatabase, and populates it with all of the blank
//         tables it needs for Modflow Analsyt.  You need to set the filename
//         before calling this.
// *****************************************************************************
HRESULT CCreateGDB::SetupNewGDB (bool a_CreateAllModflowTables/*false*/)
{
  HRESULT  hr;
  IWorkspaceFactoryPtr ipWF;
  IWorkspacePtr ipWorkspace;

  // Create a new workspace factory and workspace
  hr = CreateWorkspaceFactory(&ipWF);
  if (FAILED(hr)) return hr;
  hr = CreateWorkspace(ipWF, &ipWorkspace);
  if (FAILED(hr)) return hr;

  // Start editing
  hr = StartEditing(ipWorkspace);
  if (FAILED(hr)) return hr;

  // Create the workspace domains
  //CMDMDomains domains;
  //hr = domains.CreateAHGWDomains(ipWorkspace);
  hr = CMDMDomains::Instance().CreateAHGWDomains(ipWorkspace);

  // Add all the tables
  if (a_CreateAllModflowTables)
    hr = AddAllModflowTables(ipWorkspace);
  else
  {
    // Make some of the tables:  Output, Parameters and Observation
    hr = AddAllOuputputTables(ipWorkspace);
    if (FAILED(hr)) return hr;
    hr = AddAllParamTables(ipWorkspace);
    if (FAILED(hr)) return hr;
    hr = AddAllObservationTables(ipWorkspace);
    if (FAILED(hr)) return hr;
  }
  if (FAILED(hr)) return hr;

  // We're done, stop editing
  hr = StopEditing(ipWorkspace);
  return hr;
}

// *****************************************************************************
// NOTES:  Function for setting the path, filename, and file type for the
//         geodatabase, if we're only give the full path and filename as
//         one string.
// *****************************************************************************
void CCreateGDB::SetGDBFile2 (CComBSTR a_pathandfile)
{
  CComBSTR extension, path, file;
  int i, len, period, slash, type;

  extension = "";
  path = "";
  file = "";
  len = a_pathandfile.Length();
  period = -1;
  slash = -1;
  type = 0;

  // Loop backwards, finding the last period (file extension) and last
  // slash (for path and filename)
  for (i = len-1; i >= 0; --i) {
    if (a_pathandfile[i] == '.' &&
        period == -1)
      period = i;
    if (a_pathandfile[i] == '\\' &&
        slash == -1)
      slash = i;
  }

  // See if we found the period (for file extension and geodatabase type)
  if (period != -1) {
    // Add whatever comes after the . in the filename
    for (i = period + 1; i < len; ++i)
      extension.Append(a_pathandfile[i]);

    if (extension == "mdb" || extension == "MDB")
      type = 1;
    else if (extension == "gdb" || extension == "GDB")
      type = 0;
  }
  m_Personal = type;

  // See if we found the slash (for the file path and name)
  if (slash != -1) {
    // Add whatever comes before the slash to be the path
    for (i = 0; i < slash; ++i)
      path.Append(a_pathandfile[i]);
    // Add whatever comes after the slash to be the filename
    for (i = slash + 1; i < len; ++i)
      file.Append(a_pathandfile[i]);
  }
  m_szPath = path.Copy();
  m_szFile = file.Copy();
}

// *****************************************************************************
// NOTES:  Creates every single table we've defined below
// *****************************************************************************
HRESULT CCreateGDB::AddAllModflowTables (IFeatureWorkspacePtr ipFW)
{
  HRESULT    hr;
  ITablePtr  ipTable;
  ma_tables  tableid;
  int        i;

  // Add all tables (first is MDFGLOBALS_TABLE, last is OUTPUTSSFLOW_TABLE):
  for (i = (int)MDFGLOBALS_TABLE; i < (int)MATABLES_END; ++i) {
    tableid = (ma_tables)i;
    hr = AddTable(tableid, ipFW, &ipTable);
    if (FAILED(hr))
      return hr;
  }

  return hr;
}

// *****************************************************************************
// NOTES:  Creates all of the Output tables (blank)
// *****************************************************************************
HRESULT CCreateGDB::AddAllOuputputTables (IFeatureWorkspacePtr ipFW)
{
  HRESULT hr;
  ITablePtr  ipTable;

  hr = AddTable(OUTPUTFILES_TABLE, ipFW, &ipTable);
  if (FAILED(hr)) return hr;
  hr = AddTable(OUTPUTTIME_TABLE, ipFW, &ipTable);
  if (FAILED(hr)) return hr;
  hr = AddTable(OUTPUTHEAD_TABLE, ipFW, &ipTable);
  if (FAILED(hr)) return hr;
  hr = AddTable(OUTPUTDRAWDOWN_TABLE, ipFW, &ipTable);
  if (FAILED(hr)) return hr;
  hr = AddTable(OUTPUTFLOW_TABLE, ipFW, &ipTable);
  if (FAILED(hr)) return hr;

  return hr;
}

// *****************************************************************************
// NOTES:  Creates all of the Parameters tables (blank)
// *****************************************************************************
HRESULT CCreateGDB::AddAllParamTables (IFeatureWorkspacePtr ipFW)
{
  HRESULT hr;
  ITablePtr  ipTable;

  hr = AddTable(PARAMS_TABLE, ipFW, &ipTable);
  if (FAILED(hr)) return hr;
  hr = AddTable(CLUSTERS_TABLE, ipFW, &ipTable);
  if (FAILED(hr)) return hr;
  hr = AddTable(IZ_TABLE, ipFW, &ipTable);
  if (FAILED(hr)) return hr;
  hr = AddTable(PILOTPTS_TABLE, ipFW, &ipTable);
  if (FAILED(hr)) return hr;
  hr = AddTable(MULTNAMES_TABLE, ipFW, &ipTable);
  if (FAILED(hr)) return hr;
  hr = AddTable(MULTIPLIERS_TABLE, ipFW, &ipTable);
  if (FAILED(hr)) return hr;
  hr = AddTable(ZONES_TABLE, ipFW, &ipTable);
  if (FAILED(hr)) return hr;
  hr = AddTable(ZONENAMES_TABLE, ipFW, &ipTable);
  if (FAILED(hr)) return hr;
  hr = AddTable(PARINSTANCES_TABLE, ipFW, &ipTable);
  if (FAILED(hr)) return hr;
  hr = AddTable(PARINSTSP_TABLE, ipFW, &ipTable);
  if (FAILED(hr)) return hr;

  return hr;
}

// *****************************************************************************
// NOTES:  Creates all of the Observation tables (blank)
// *****************************************************************************
HRESULT CCreateGDB::AddAllObservationTables (IFeatureWorkspacePtr ipFW)
{
  HRESULT hr;
  ITablePtr  ipTable;

  hr = AddTable(OBSVARS_TABLE, ipFW, &ipTable);
  if (FAILED(hr)) return hr;
  hr = AddTable(HOB_TABLE, ipFW, &ipTable);
  if (FAILED(hr)) return hr;
  hr = AddTable(HOBLAYERS_TABLE, ipFW, &ipTable);
  if (FAILED(hr)) return hr;
  hr = AddTable(HOBTIMES_TABLE, ipFW, &ipTable);
  if (FAILED(hr)) return hr;
  hr = AddTable(FLOB_TABLE, ipFW, &ipTable);
  if (FAILED(hr)) return hr;
  hr = AddTable(FLOBFACTORS_TABLE, ipFW, &ipTable);
  if (FAILED(hr)) return hr;

  return hr;
}

// *****************************************************************************
// NOTES:  Creates a table, based on the ID passed in.  The basic field
//         information is used to create the table.
// *****************************************************************************
HRESULT CCreateGDB::AddTable (ma_tables TableID, IFeatureWorkspacePtr ipFW,
                              ITable **ppTable)
{
  HRESULT         hr;
  IFieldsEditPtr  ipFieldsEdit;
  IFieldEditPtr   ipFieldEdit;
  IUIDPtr         ipUID, ipEXTUID;
  ITablePtr       ipTable;
  CTableDef       TableDef;
  CComSafeArray<BSTR> vecNames;
  VEC_FTYPE       vecTypes;
  VEC_DOMAIN      vecDomains;
  CComBSTR        szName;

  if (!TableDef.GetData((int)TableID, vecNames, vecTypes, vecDomains, szName))
    return E_FAIL;

  // Try and open the table, if it already exists
  if (TableExists(ipFW, szName))
  {
    hr = ipFW->OpenTable(szName, &ipTable);
    if (SUCCEEDED(hr))
      return hr;
  }

  // Create a new set of basic required fields
  hr = CreateFieldsEditTable(&ipFieldsEdit);
  if (FAILED(hr))
    return hr;

  // Now add the fields for the MDF_Models table
  hr = AddModflowFieldsToFieldsEdit(TableID, ipFieldsEdit);

  // Now, create the table
  hr = CreateTableUID(&ipUID, &ipEXTUID);
  if (FAILED(hr)) return hr;
  hr = CreateNewTable(szName, ipFW, ipFieldsEdit,
                      ipUID, ipEXTUID, &ipTable);

  return hr;
}

// *****************************************************************************
// NOTES:  Loops through the tables in the workspace, and tries to find the
//         table with the name passed in
// *****************************************************************************
bool CCreateGDB::TableExists (IFeatureWorkspacePtr ipFW, CComBSTR szName)
{
  HRESULT hr;
  IWorkspacePtr ipWS(ipFW);
  IEnumDatasetNamePtr ipEnumDatasetName;
  IDatasetNamePtr ipDatasetName;
  CComBSTR bstrName;

  if (ipWS)
  {
    hr = ipWS->get_DatasetNames(esriDTTable, &ipEnumDatasetName);
    if (FAILED(hr)) return false;

    if (ipEnumDatasetName)
    {
      hr = ipEnumDatasetName->Next(&ipDatasetName);
      if (FAILED(hr)) return false;

      while (ipDatasetName)
      {
        hr = ipDatasetName->get_Name(&bstrName);
        if (FAILED(hr)) return false;

        if (bstrName == szName)
          return true;

        hr = ipEnumDatasetName->Next(&ipDatasetName);
        if (FAILED(hr)) return false;
      }
    }
  }

  return false;
}

// *****************************************************************************
// NOTES:  Given a table type in Modflow Analyst, add the additional fields
//         needed by that table.
// *****************************************************************************
HRESULT CCreateGDB::AddModflowFieldsToFieldsEdit (ma_tables ftype,
                                                  IFieldsEditPtr ipFE)
{
  try
  {
    int            i;
    HRESULT        hr = S_OK;
    IFieldEditPtr  ipFieldEdit;
    CComSafeArray<BSTR>       vecNames;
    VEC_FTYPE      vecTypes;
    VEC_DOMAIN     vecDomains;
    CTableDef      TableDef;
    CComBSTR       szTableName;

    // Get the field names and field types for this table type
    if (!TableDef.GetData((int)ftype, vecNames, vecTypes, vecDomains, szTableName))
      return E_FAIL;

    for (i = 0; i < (int)vecNames.GetCount(); ++i) {
      hr = CreateField(vecNames.GetAt(i), vecTypes[i], true, -1, 500, -1, -1,
                       (vecDomains.empty() ? NULL : vecDomains.at(i)), &ipFieldEdit);
      if (FAILED(hr)) return hr;

      hr = ipFE->AddField(ipFieldEdit);
      if (FAILED(hr)) return hr;
    }

    return hr;
  }
  catch (std::exception)
  {
    return E_FAIL;
  }
}

// *****************************************************************************
// NOTES:  Adds Cell2D features (polygons) for the given instance to the
//         feature class you pass in.
// *****************************************************************************
HRESULT CCreateGDB::AddGridFeatures (IFeatureClassPtr ipFC, int Instance)
{
  HRESULT                  hr;
  IFeatureCursorPtr        ipFCursor;
  IFeatureBufferPtr        ipFBuffer;
  IPolygonPtr              ipGrid;
  IPointPtr                ipPoint1, ipPoint2, ipPoint3, ipPoint4;
  IZAwarePtr               ipZAware;
  IPointCollectionPtr      ipPointCollection;
  IGeometryCollectionPtr   ipGeometryCollection;
  IRingPtr                 ipRing;
  IGeometryPtr             ipGeometry;
  ITopologicalOperatorPtr  ipOperator;
  CComVariant              varID;

  // Make sure we've got a valid feature class first
  if (ipFC == 0)
    return E_FAIL;

  // Get a feature cursor and feature buffer for inserting new features
  hr = ipFC->Insert(VARIANT_TRUE, &ipFCursor);
  if (FAILED(hr)) return hr;
  hr = ipFC->CreateFeatureBuffer(&ipFBuffer);
  if (FAILED(hr)) return hr;

  // Create some of the objects we need to define the grid cells
  hr = ipPoint1.CreateInstance(CLSID_Point);
  if (FAILED(hr)) return hr;
  hr = ipPoint2.CreateInstance(CLSID_Point);
  if (FAILED(hr)) return hr;
  hr = ipPoint3.CreateInstance(CLSID_Point);
  if (FAILED(hr)) return hr;
  hr = ipPoint4.CreateInstance(CLSID_Point);
  if (FAILED(hr)) return hr;
  hr = ipPointCollection.CreateInstance(CLSID_Ring);
  if (FAILED(hr)) return hr;
  hr = ipGeometryCollection.CreateInstance(CLSID_Polygon);
  if (FAILED(hr)) return hr;

  // Set the ZAware property of some of these objects
  ipZAware = ipPoint1;
  hr = ipZAware->put_ZAware(VARIANT_TRUE);
  if (FAILED(hr)) return hr;
  ipZAware = ipPoint2;
  hr = ipZAware->put_ZAware(VARIANT_TRUE);
  if (FAILED(hr)) return hr;
  ipZAware = ipPoint3;
  hr = ipZAware->put_ZAware(VARIANT_TRUE);
  if (FAILED(hr)) return hr;
  ipZAware = ipPoint4;
  hr = ipZAware->put_ZAware(VARIANT_TRUE);
  if (FAILED(hr)) return hr;
  ipZAware = ipGeometryCollection;
  hr = ipZAware->put_ZAware(VARIANT_TRUE);
  if (FAILED(hr)) return hr;

  {

    // DEBUG:
    // Add the corner points of the grid cell to the point collection
    hr = ipPoint1->PutCoords(00.0, 00.0); hr = ipPoint1->put_Z(0.0);
    hr = ipPointCollection->AddPoint(ipPoint1);
    hr = ipPoint2->PutCoords(10.0, 00.0); hr = ipPoint2->put_Z(0.0);
    hr = ipPointCollection->AddPoint(ipPoint2);
    hr = ipPoint3->PutCoords(10.0, 10.0); hr = ipPoint3->put_Z(0.0);
    hr = ipPointCollection->AddPoint(ipPoint3);
    hr = ipPoint4->PutCoords(00.0, 10.0); hr = ipPoint4->put_Z(0.0);
    hr = ipPointCollection->AddPoint(ipPoint4);

    // Take the point collection, store it in a ring and geometry
    ipRing = ipPointCollection;
    ipGeometry = ipRing;
    hr = ipGeometryCollection->AddGeometry(ipGeometry);

    // Put the geometry into a polygon and simplify
    ipGrid = ipGeometryCollection;
    ipOperator = ipGrid;
    hr = ipOperator->Simplify();
    if (FAILED(hr)) return hr;

    // Store the grid cell in the feature buffer
    hr = ipFBuffer->putref_Shape(ipGrid);
    if (FAILED(hr)) return hr;
    // .... Set any field values if necessary

    // Store the feature buffer into the feature cursor
    hr = ipFCursor->InsertFeature(ipFBuffer, &varID);
    if (FAILED(hr)) return hr;

  }

  return hr;
}

// *****************************************************************************
// NOTES:  Adds Node features (points) for the given instance to the
//         feature class you pass in.
// *****************************************************************************
HRESULT CCreateGDB::AddNodeFeatures (IFeatureClassPtr ipFC, int Instance)
{
  HRESULT            hr;
  IFeatureCursorPtr  ipFCursor;
  IFeatureBufferPtr  ipFBuffer;
  IPointPtr          ipPoint;
  IZAwarePtr         ipZAware;
  CComVariant        varID;

  // Make sure we've got a valid feature class first
  if (ipFC == 0)
    return E_FAIL;

  // Get a feature cursor and feature buffer for inserting new features
  hr = ipFC->Insert(VARIANT_TRUE, &ipFCursor);
  if (FAILED(hr)) return hr;
  hr = ipFC->CreateFeatureBuffer(&ipFBuffer);
  if (FAILED(hr)) return hr;

  // Create a new point
  hr = ipPoint.CreateInstance(CLSID_Point);
  if (FAILED(hr)) return hr;

  // Set up the ZAware attributes for the point
  ipZAware = ipPoint;
  hr = ipZAware->put_ZAware(VARIANT_TRUE);
  if (FAILED(hr)) return hr;

  {

    // Put the coordinates in the point
    hr = ipPoint->PutCoords(5.0, 5.0);
    if (FAILED(hr)) return hr;
    hr = ipPoint->put_Z(5.0);
    if (FAILED(hr)) return hr;

    // Store the node in the feature buffer
    hr = ipFBuffer->putref_Shape(ipPoint);
    if (FAILED(hr)) return hr;
    // .... Set any field values if necessary

    // Store the feature buffer into the feature cursor
    hr = ipFCursor->InsertFeature(ipFBuffer, &varID);
    if (FAILED(hr)) return hr;
  }

  return hr;
}

// *****************************************************************************
// NOTES:  For debugging only..... delete this function later
// *****************************************************************************
void CCreateGDB::DEBUG_DELETE (IFeatureWorkspacePtr ipFW)
{
  HRESULT           hr;
  IUIDPtr           ipUID;
  IFeatureClassPtr  ipFeatureClassGrid;
  IFeatureClassPtr  ipFeatureClassNode;

  hr = CreateFeatureUID(&ipUID);
  if (FAILED(hr)) return;

  // =================================================================
  // Create a feature class for adding grid stuff
  hr = CreateBasicFeatureClassWksp(L"Grid", esriGeometryPolygon,
                                   ipFW, ipUID, &ipFeatureClassGrid);
  if (FAILED(hr)) return;

  // Now, add polygon features to it
  hr = AddGridFeatures(ipFeatureClassGrid, 0);
  if (FAILED(hr)) return;


  // =================================================================
  // Create a feature class for adding node stuff
  hr = CreateBasicFeatureClassWksp(L"Node", esriGeometryPoint,
                                   ipFW, ipUID, &ipFeatureClassNode);
  if (FAILED(hr)) return;

  // Now, add polygon features to it
  hr = AddNodeFeatures(ipFeatureClassNode, 0);
  if (FAILED(hr)) return;

}

// *****************************************************************************
// NOTES:  Checks if a file exists
// *****************************************************************************
bool CCreateGDB::FileExists (const std::string& strFilename)
{
  struct stat stFileInfo;
  bool blnReturn;
  int intStat;

  // Attempt to get the file attributes
  intStat = stat(strFilename.c_str(), &stFileInfo);

  if (intStat == 0) {
    // We were able to get the file attributes
    // so the file obviously exists.
    blnReturn = true;
  }
  else {
    // We were not able to get the file attributes.
    // This may mean that we don't have permission to
    // access the folder which contains this file. If you
    // need to do that level of checking, lookup the
    // return values of stat which will give you
    // more details on why stat failed.
    blnReturn = false;
  }
  
  return (blnReturn); 
} // CCreateGDB::FileExists